import k3d
import numpy as np
import pytest
from .plot_compare import *
from k3d.helpers import download
import vtk
from vtk.util import numpy_support

vertices = [
    -10, 0, -1,
    10, 0, -1,
    10, 0, 1,
    -10, 0, 1,
]

indices = [
    0, 1, 3,
    1, 2, 3
]


def test_mesh():
    global vertices, indices

    prepare()

    mesh = k3d.mesh(vertices, indices)
    pytest.plot += mesh

    compare('mesh')


def test_mesh_attribute():
    global vertices, indices

    prepare()

    vertex_attribute = [0, 1, 1, 0]
    mesh = k3d.mesh(vertices, indices, attribute=vertex_attribute,
                    color_map=k3d.basic_color_maps.CoolWarm, color_range=[0.0, 1.0])
    pytest.plot += mesh

    compare('mesh_attribute')


def test_mesh_advanced():
    prepare()

    filename = download(
        'https://github.com/To-Fujita/Babylon.js_3D_Graphics/raw/master/scenes/stl/Cute%20Darth%20Vader.stl')

    reader = vtk.vtkSTLReader()
    reader.SetFileName(filename)
    reader.Update()

    mesh = k3d.vtk_poly_data(reader.GetOutput(), color=0x222222, flat_shading=True,
                             transform=k3d.transform(rotation=[np.pi / 2, 1, 0, 0]))
    pytest.plot += mesh

    compare('mesh_advanced')


def test_mesh_advanced_smoothed():
    prepare()

    filename = download(
        'https://github.com/To-Fujita/Babylon.js_3D_Graphics/raw/master/scenes/stl/Cute%20Darth%20Vader.stl')

    reader = vtk.vtkSTLReader()
    reader.SetFileName(filename)
    reader.Update()

    mesh = k3d.vtk_poly_data(reader.GetOutput(), color=0x222222, flat_shading=False,
                             transform=k3d.transform(rotation=[np.pi / 2, 1, 0, 0]))
    pytest.plot += mesh

    compare('mesh_advanced_smoothed')


def test_mesh_advanced_opacity():
    prepare()

    filename = download(
        'https://github.com/To-Fujita/Babylon.js_3D_Graphics/raw/master/scenes/stl/Cute%20Darth%20Vader.stl')

    reader = vtk.vtkSTLReader()
    reader.SetFileName(filename)
    reader.Update()

    mesh = k3d.vtk_poly_data(reader.GetOutput(), color=0x222222, flat_shading=False, opacity=0.5,
                             transform=k3d.transform(rotation=[np.pi / 2, 1, 0, 0]))
    pytest.plot += mesh

    compare('mesh_advanced_opacity')


def test_mesh_advanced_wireframe():
    prepare()

    filename = download(
        'https://github.com/To-Fujita/Babylon.js_3D_Graphics/raw/master/scenes/stl/Cute%20Darth%20Vader.stl')

    reader = vtk.vtkSTLReader()
    reader.SetFileName(filename)
    reader.Update()

    mesh = k3d.vtk_poly_data(reader.GetOutput(), color=0x222222, opacity=0.2, wireframe=True,
                             transform=k3d.transform(rotation=[np.pi / 2, 1, 0, 0]))
    pytest.plot += mesh

    compare('mesh_advanced_wireframe')


def test_mesh_attribute_advanced():
    prepare()

    N = 100

    theta = np.linspace(0, 2.0 * np.pi, N)
    phi = np.linspace(0, 2.0 * np.pi, N)
    theta, phi = np.meshgrid(theta, phi)

    c, a = 2, 1
    x = (c + a * np.cos(theta)) * np.cos(phi)
    y = (c + a * np.cos(theta)) * np.sin(phi)
    z = a * np.sin(theta)

    vertices = np.dstack([x, y, z]).astype(np.float32)
    indices = (np.stack([
        np.arange(N * N) + 0, np.arange(N * N) + N, np.arange(N * N) + N + 1,
        np.arange(N * N) + 0, np.arange(N * N) + N + 1, np.arange(N * N) + 1
    ]).T % (N * N)).astype(np.uint32)

    mesh = k3d.mesh(vertices, indices, flat_shading=False,
                    attribute=phi,
                    transform=k3d.transform(rotation=[np.pi / 2, 1, 0, 0]),
                    color_map=k3d.matplotlib_color_maps.twilight)

    pytest.plot += mesh

    compare('mesh_attribute_advanced')

    pytest.plot.clipping_planes = [
        [1, 1, 0, 0]
    ]

    compare('mesh_attribute_advanced_clipping_planes')


def test_mesh_triangle_attribute():
    prepare()

    filename = download(
        'https://github.com/To-Fujita/Babylon.js_3D_Graphics/raw/master/scenes/stl/Cute%20Darth%20Vader.stl')

    reader = vtk.vtkSTLReader()
    reader.SetFileName(filename)
    reader.Update()

    qualityFilter = vtk.vtkMeshQuality()
    qualityFilter.SetInputData(reader.GetOutput())
    qualityFilter.SetTriangleQualityMeasureToArea()
    qualityFilter.SetQuadQualityMeasureToArea()
    qualityFilter.Update()

    mesh = k3d.vtk_poly_data(qualityFilter.GetOutput(), cell_color_attribute=('Quality', 0.0, 0.83),
                             transform=k3d.transform(rotation=[np.pi / 2, 1, 0, 0]))

    pytest.plot += mesh

    compare('mesh_triangle_attribute')


def test_mesh_volume_data():
    prepare()

    filename = download(
        'https://github.com/To-Fujita/Babylon.js_3D_Graphics/raw/master/scenes/stl/Cute%20Darth%20Vader.stl')

    reader = vtk.vtkSTLReader()
    reader.SetFileName(filename)
    reader.Update()
    poly = reader.GetOutput()

    reader = vtk.vtkXMLImageDataReader()
    reader.SetFileName('./test/assets/volume.vti')
    reader.Update()
    vti = reader.GetOutput()

    x, y, z = vti.GetDimensions()
    volume_data = numpy_support.vtk_to_numpy(
        vti.GetPointData().GetArray(0)
    ).reshape(-1, y, x).astype(np.float32)

    mesh = k3d.vtk_poly_data(poly, color=0xffffff, volume=volume_data,
                             transform=k3d.transform(rotation=[np.pi / 2, 1, 0, 0]),
                             volume_bounds=[-50, 150, -200, 100, -50, 250])

    pytest.plot += mesh

    compare('mesh_volume_data')



import threading
import numpy as np
import copy
import msgpack
from flask import Flask, request, Response, send_from_directory
from base64 import b64decode
import logging
import atexit
import urllib.request
import time

from .helpers import to_json

log = logging.getLogger('werkzeug')
log.setLevel(logging.ERROR)


# logging.basicConfig(filename='test.log', level=logging.DEBUG)


class k3d_remote:
    def __init__(self, k3d_plot, driver, width=1280, height=720, port=8080):

        driver.set_window_size(width, height)

        self.port = port
        self.browser = driver
        self.k3d_plot = k3d_plot

        self.api = Flask(__name__)

        thread = threading.Thread(target=lambda: self.api.run(host="0.0.0.0", port=port),
                                  daemon=True)
        thread.deamon = True
        thread.start()

        self.synced_plot = {k: None for k in k3d_plot.get_plot_params().keys()}
        self.synced_objects = {}

        @self.api.route('/stop', methods=['GET'])
        def stopServer():
            shutdown_hook = request.environ.get('werkzeug.server.shutdown')

            if shutdown_hook is not None:
                shutdown_hook()

            return Response("Bye", mimetype='text/plain')

        @self.api.route('/<path:path>')
        def static_file(path):
            root_dir = self.k3d_plot.get_static_path()
            return send_from_directory(root_dir, path)

        @self.api.route('/ping')
        def ping():
            return Response(":)")

        @self.api.route('/', methods=['POST'])
        def generate():
            current_plot_params = self.k3d_plot.get_plot_params()
            plot_diff = {k: current_plot_params[k] for k in current_plot_params.keys()
                         if current_plot_params[k] != self.synced_plot[k]}

            objects_diff = {}

            for o in self.k3d_plot.objects:
                if o.id not in self.synced_objects:
                    objects_diff[o.id] = {k: to_json(k, o[k], o) for k in o.keys if
                                          not k.startswith('_')}
                else:
                    for p in o.keys:
                        if p.startswith('_'):
                            continue

                        sync = False

                        if isinstance(o[p], np.ndarray):
                            sync = (o[p] != self.synced_objects[o.id][p]).any()
                        else:
                            sync = o[p] != self.synced_objects[o.id][p]

                        if sync:
                            if o.id not in objects_diff.keys():
                                objects_diff[o.id] = {"id": o.id, "type": o.type}

                            objects_diff[o.id][p] = to_json(p, o[p], o)

            for k in self.synced_objects.keys():
                if k not in self.k3d_plot.object_ids:
                    objects_diff[k] = None  # to remove from plot

            diff = {
                "plot_diff": plot_diff,
                "objects_diff": objects_diff
            }

            self.synced_objects = {v.id: {k: copy.deepcopy(v[k]) for k in v.keys} for v in
                                   self.k3d_plot.objects}
            self.synced_plot = current_plot_params

            return Response(msgpack.packb(diff, use_bin_type=True),
                            mimetype='application/octet-stream')

        self.browser.implicitly_wait(5)
        self.browser.get(url="http://localhost:" + str(port) + "/headless.html")

        atexit.register(self.close)

    def sync(self, hold_until_refreshed=False):
        self.browser.execute_script("k3dRefresh()")

        if hold_until_refreshed:
            while self.browser.execute_script("return window.refreshed") == False:
                time.sleep(0.1)

    def get_browser_screenshot(self):
        return self.browser.get_screenshot_as_png()

    def camera_reset(self, factor=1.5):
        self.browser.execute_script("K3DInstance.resetCamera(%f)" % factor)
        # refresh dom elements
        self.browser.execute_script("K3DInstance.refreshGrid()")
        self.browser.execute_script("K3DInstance.dispatch(K3DInstance.events.RENDERED)")

    def get_screenshot(self, only_canvas=False):
        screenshot = self.browser.execute_script("""
        return K3DInstance.getScreenshot(K3DInstance.parameters.screenshotScale, %d).then(function (d){
        return d.toDataURL().split(',')[1];
        });                                 
        """ % only_canvas)

        return b64decode(screenshot)

    def close(self):
        if self.api is not None:
            urllib.request.urlopen("http://localhost:" + str(self.port) + "/stop")
            self.api = None

        if self.browser is not None:
            self.browser.close()
            self.browser = None


def get_headless_driver(no_headless=False):
    from selenium.webdriver.chrome.options import Options
    from selenium import webdriver
    from webdriver_manager.chrome import ChromeDriverManager

    options = Options()

    if not no_headless:
        options.add_argument("--headless")

    return webdriver.Chrome(ChromeDriverManager().install(), options=options)



import k3d
import numpy as np
import pytest
from .plot_compare import *
from k3d.helpers import download


def test_stl():
    prepare()

    filename = download(
        'https://github.com/To-Fujita/Babylon.js_3D_Graphics/raw/master/scenes/stl/Cute%20Darth%20Vader.stl')

    mesh = k3d.stl(open(filename, 'rb').read(), color=0x222222, flat_shading=True,
                   transform=k3d.transform(rotation=[np.pi / 2, 1, 0, 0]))

    pytest.plot += mesh

    compare('stl')


def test_stl_color():
    prepare()

    filename = download(
        'https://github.com/To-Fujita/Babylon.js_3D_Graphics/raw/master/scenes/stl/Cute%20Darth%20Vader.stl')

    mesh = k3d.stl(open(filename, 'rb').read(), color=0xff00ff,
                   transform=k3d.transform(rotation=[np.pi / 2, 1, 0, 0]))

    pytest.plot += mesh

    compare('stl_color')


def test_stl_wireframe():
    prepare()

    filename = download(
        'https://github.com/To-Fujita/Babylon.js_3D_Graphics/raw/master/scenes/stl/Cute%20Darth%20Vader.stl')

    mesh = k3d.stl(open(filename, 'rb').read(), wireframe=True,
                   transform=k3d.transform(rotation=[np.pi / 2, 1, 0, 0]))

    pytest.plot += mesh

    compare('stl_wireframe')


def test_stl_smooth():
    prepare()

    filename = download(
        'https://github.com/To-Fujita/Babylon.js_3D_Graphics/raw/master/scenes/stl/Cute%20Darth%20Vader.stl')

    mesh = k3d.stl(open(filename, 'rb').read(), flat_shading=False, color=0x222222,
                   transform=k3d.transform(rotation=[np.pi / 2, 1, 0, 0]))

    pytest.plot += mesh

    compare('stl_smooth')



"""
Utilities module.
"""

from typing import Sequence, Tuple
import itertools
import os
import zlib

import numpy as np

# import logging
# from pprint import pprint, pformat
#
# logger = logging.getLogger()
# logger.setLevel(logging.INFO)
# fh = logging.FileHandler('k3d.log')
# fh.setLevel(logging.INFO)
# logger.addHandler(fh)


# pylint: disable=unused-argument
# noinspection PyUnusedLocal
def array_to_binary(ar, compression_level=0, force_contiguous=True):
    """Pre-process numpy array for serialization in traittypes.Array."""
    if ar.dtype.kind not in ["u", "i", "f"]:  # ints and floats
        raise ValueError("unsupported dtype: %s" % ar.dtype)

    if ar.dtype == np.float64:  # WebGL does not support float64
        ar = ar.astype(np.float32)
    elif ar.dtype == np.int64:  # JS does not support int64
        ar = ar.astype(np.int32)

    if force_contiguous and not ar.flags["C_CONTIGUOUS"]:  # make sure it's contiguous
        ar = np.ascontiguousarray(ar)

    if compression_level > 0:
        return {
            "compressed_data": zlib.compress(ar.flatten(), compression_level),
            "dtype": str(ar.dtype),
            "shape": ar.shape,
        }
    else:
        return {
            "data": memoryview(ar.flatten()),
            "dtype": str(ar.dtype),
            "shape": ar.shape,
        }


# noinspection PyUnusedLocal
def from_json_to_array(value, obj=None):
    """Post-process traittypes.Array after deserialization to numpy array."""
    if value:
        if "data" in value:
            return np.frombuffer(value["data"], dtype=value["dtype"]).reshape(
                value["shape"]
            )
        else:
            return np.frombuffer(
                zlib.decompress(value["compressed_data"]), dtype=value["dtype"]
            ).reshape(value["shape"])
    return None


def to_json(name, input, obj=None, compression_level=0):
    if hasattr(obj, "compression_level"):
        compression_level = obj.compression_level

    if isinstance(input, dict):
        property = obj[name]
        ret = {}
        for key, value in input.items():
            ret[str(key)] = to_json(key, value, property, compression_level)

        return ret
    elif isinstance(input, np.ndarray) and input.dtype is np.dtype(object):
        return to_json(name, input.tolist(), obj, compression_level)
    elif isinstance(input, list):
        property = obj[name]
        return [
            to_json(idx, v, property, compression_level) for idx, v in enumerate(input)
        ]
    elif isinstance(input, bytes):
        return array_to_binary(np.frombuffer(input, dtype=np.uint8), compression_level)
    elif isinstance(input, np.ndarray):
        return array_to_binary(input, compression_level)
    else:
        return input


def from_json(input, obj=None):
    # logger.info('from_json:' + pformat(input))

    if (
            isinstance(input, dict)
            and "dtype" in input
            and ("data" in input or "compressed_data" in input)
            and "shape" in input
    ):
        return from_json_to_array(input, obj)
    elif isinstance(input, list):
        return [from_json(i, obj) for i in input]
    elif isinstance(input, dict):
        ret = {}
        for key, value in input.items():
            ret[key] = from_json(value, obj)

        return ret
    else:
        return input


def array_serialization_wrap(name):
    return {
        "to_json": (lambda input, obj: to_json(name, input, obj)),
        "from_json": from_json,
    }


def callback_serialization_wrap(name):
    return {
        "to_json": (lambda input, obj: obj[name] is not None),
        "from_json": from_json,
    }


def download(url) -> str:
    """Retrieve the file at url, save it locally and return the path."""
    basename = os.path.basename(url)
    if os.path.exists(basename):
        return basename

    from urllib.request import urlopen

    print("Downloading: {}".format(basename))

    with urlopen(url) as response, open(basename, "wb") as output:
        output.write(response.read())

    return basename


def minmax(iterable):
    """Return [min(iterable), max(iterable)].

    This should be a built in function in Python, and has even been proposed on Python-ideas newsgroup.
    This is not to be confused with the algorithm for finding winning strategies in 2-player games."""
    return [float(np.nanmin(iterable)), float(np.nanmax(iterable))]


def check_attribute_range(attribute, color_range=()) -> Tuple[float, float]:
    """Provide color range versus provided attribute, compute color range if necessary.

    If the attribute is empty or color_range has 2 elements, returns color_range unchanged.
    Computes color range as [min(attribute), max(attribute)].
    When min(attribute) == max(attribute) returns [min(attribute), min(attribute)+1]."""
    if type(attribute) is dict or attribute.size == 0 or len(color_range) == 2:
        return color_range
    color_range = minmax(attribute)
    if color_range[0] == color_range[1]:
        color_range[1] += 1.0
    return color_range


def map_colors(attribute, color_map, color_range=()) -> Sequence[int]:
    a_min, a_max = check_attribute_range(attribute, color_range)
    map_array = np.asarray(color_map)
    map_array = map_array.reshape((map_array.size // 4, 4))
    attribute = (attribute - a_min) / (
            a_max - a_min
    )  # normalizing attribute for range lookup
    red, green, blue = [
        np.array(
            255 * np.interp(attribute, xp=map_array[:, 0], fp=map_array[:, i + 1]),
            dtype=np.int32,
        )
        for i in range(3)
    ]
    colors = (red << 16) + (green << 8) + blue
    return colors


def bounding_corners(bounds, z_bounds=(0.0, 1)):
    """Return corner point coordinates for bounds array."""
    return np.array(
        list(itertools.product(bounds[:2], bounds[2:4], bounds[4:] or z_bounds))
    )


def min_bounding_dimension(bounds):
    """Return a minimal dimension along axis in a bounds ([min_x, max_x, min_y, max_y, min_z, max_z]) array."""
    return min(abs(x1 - x0) for x0, x1 in zip(bounds, bounds[1:]))


def shape_validation(*dimensions):
    """Create a validator callback (for Array traittype) ensuring shape."""
    from traitlets import TraitError

    def validator(trait, value):
        if np.shape(value) != dimensions:
            raise TraitError(
                "Expected an array of shape %s and got %s" % (dimensions, value.shape)
            )

        return value

    return validator


def validate_sparse_voxels(trait, value):
    """Check sparse voxels for array shape and values."""
    from traitlets import TraitError

    if len(value.shape) != 2 or value.shape[1] != 4:
        raise TraitError(
            "Expected an array of shape (N, 4) and got %s" % (value.shape,)
        )

    if (value.astype(np.int16) < 0).any():
        raise TraitError("Voxel coordinates and values must be non-negative")

    return value


def quad(w, h):
    w /= 2
    h /= 2
    vertices = np.array([-w, -h, -0, w, -h, 0, w, h, 0, -w, h, 0], dtype=np.float32)
    indices = np.array([0, 1, 2, 0, 2, 3], dtype=np.uint32)

    return vertices, indices


def get_bounding_box(model_matrix, boundary=[-0.5, 0.5, -0.5, 0.5, -0.5, 0.5]):
    b_min = np.array([boundary[0], boundary[2], boundary[4], 0])
    b_max = np.array([boundary[1], boundary[3], boundary[5], 0])

    b_min = model_matrix.dot(b_min)
    b_max = model_matrix.dot(b_max)

    return np.dstack([b_min[0:3], b_max[0:3]]).flatten()


def get_bounding_box_points(arr, model_matrix):
    d = arr.flatten()

    if d.shape[0] < 3:
        d = np.array([0, 0, 0])

    # fmt: off
    boundary = np.array([
        np.min(d[0::3]), np.max(d[0::3]),
        np.min(d[1::3]), np.max(d[1::3]),
        np.min(d[2::3]), np.max(d[2::3]),
    ])
    # fmt: on

    return get_bounding_box(model_matrix, boundary)


def get_bounding_box_point(position):
    return np.dstack([np.array(position), np.array(position)]).flatten()



import k3d
import pytest
from .plot_compare import *


def test_labels():
    prepare()

    pytest.plot += k3d.label('Test dynamic', (0, 0, 0), mode='dynamic')
    pytest.plot += k3d.label('Test local', (1.5, 0, 0), mode='local')
    pytest.plot += k3d.label('Test side 1', (1.5, 1.5, 1.5), mode='side')
    pytest.plot += k3d.label('Test side 2', (0, 0, 1.5), mode='side', is_html=True)

    compare('labels', False)


def test_labels_without_box():
    prepare()

    pytest.plot += k3d.label('Test dynamic', (0, 0, 0), mode='dynamic', label_box=False)
    pytest.plot += k3d.label('Test local', (1.5, 0, 0), mode='local', label_box=False)
    pytest.plot += k3d.label('Test side 1', (1.5, 1.5, 1.5), mode='side', label_box=False)
    pytest.plot += k3d.label('Test side 2', (0, 0, 1.5), mode='side', is_html=True, label_box=False)

    compare('labels_without_box', False)



import k3d
import pytest
from .plot_compare import *
import numpy as np


def test_vectors():
    prepare()

    origins = [1.0, 1.0, 0.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    vectors = [1.0, 0.0, 0.0, 1.0, 1.0, 0.0, 1.0, 1.0, 1.0]
    colors = [0xff0000, 0x0000ff, 0x0000ff, 0xff0000, 0x0000ff, 0xff0000]

    vectors = k3d.vectors(origins, vectors, colors=colors, labels=[], label_size=1.5)
    pytest.plot += vectors

    compare('vectors')


def test_vectors_labels():
    prepare()

    origins = [1.2, 1.2, 0.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    vectors = [1.5, 0.0, 0.0, 1.5, 1.5, 0.0, 1.5, 1.5, 1.5]
    colors = [0xff0000, 0x0000ff, 0x0000ff, 0xff0000, 0x0000ff, 0xff0000]

    vectors = k3d.vectors(origins, vectors, colors=colors, label_size=1.5,
                          labels=['aa', 'bb', 'cc'])
    pytest.plot += vectors

    compare('vectors_labels', False)

    vectors.head_size = 3.0

    compare('vectors_labels_dynamic_head_size', False)


def test_vectors_advance():
    prepare()

    n = 20
    x = np.linspace(-5, 5, n)
    y = np.linspace(-5, 5, n)
    z = np.linspace(-5, 5, n)
    xx, yy, zz = np.meshgrid(x, y, z)
    uu, vv, ww = zz, yy, xx
    xx, yy, zz, uu, vv, ww = [t.flatten().astype(np.float32) for t in [xx, yy, zz, uu, vv, ww]]
    scale = 0.25
    magnitude = np.sqrt(uu ** 2 + vv ** 2 + ww ** 2)
    vectors = np.array((uu, vv, ww)).T * scale
    origins = np.array((xx, yy, zz)).T
    colors = k3d.helpers.map_colors(magnitude, k3d.matplotlib_color_maps.Plasma, [])
    vec_colors = np.zeros(2 * len(colors))
    for i, c in enumerate(colors):
        vec_colors[2 * i] = c
        vec_colors[2 * i + 1] = c
    vec_colors = vec_colors.astype(np.uint32)
    vector_field = k3d.vectors(
        origins=origins - vectors / 2,
        vectors=vectors,
        colors=vec_colors,
    )

    pytest.plot += vector_field

    compare('vectors_advance')



import unittest

from traitlets import TraitError
import numpy as np

from ..objects import Drawable
from ..factory import text


class TestDrawable(unittest.TestCase):
    def setUp(self):
        self.obj = Drawable()

    def test_iteration_returns_object(self):
        for obj in self.obj:
            self.assertTrue(isinstance(obj, Drawable))

    def test_can_add_another_object(self):
        self.obj += Drawable()

        for obj in self.obj:
            self.assertTrue(isinstance(obj, Drawable))

    def test_can_add_many_objects(self):
        obj = Drawable()
        self.obj += obj + obj

        for obj in self.obj:
            self.assertTrue(isinstance(obj, Drawable))


class TestText(unittest.TestCase):
    def test_position_accepts_numpy(self):
        text_ = text('test', [0, 0, 0])
        text_.position = np.arange(3)


class TestSTL(unittest.TestCase):
    def test_creation(self):
        from ..objects import STL
        s = STL(text='''
solid
    facet normal 0 0 0
        outer loop
            vertex -1.000000 1.000000 -1.000000
            vertex -1.000000 -1.000000 -1.000000
            vertex -1.000000 -1.000000 1.000000
        endloop
    endfacet
endsolid
        '''.strip())

    def test_creation_named(self):
        from ..objects import STL
        STL(text='''
solid named_solid
    facet normal 0 0 0
        outer loop
            vertex -1.000000 1.000000 -1.000000
            vertex -1.000000 -1.000000 -1.000000
            vertex -1.000000 -1.000000 1.000000
        endloop
    endfacet
endsolid named_solid
        '''.strip())

    def test_invalid(self):
        from ..objects import STL
        s = STL(text='''
solid
    facet normal 0 0 0
        outer loop
            vertex -1.000000 1.000000 -1.000000
            vertex -1.000000 -1.000000 -1.000000
            vertex -1.000000 -1.000000 1.000000
        endloop
    endfacet
endsolid
        '''.strip())

        def assign_bad():
            # missing endsolid, gibberish after facet normal:
            s.text = '''solid
    facet normal 0 0 0 bebebe
        outer loop
            vertex -1.000000 1.000000 -1.000000
            vertex -1.000000 -1.000000 -1.000000
            vertex -1.000000 -1.000000 1.000000
        endloop
    endfacet
        '''

        self.assertRaises(TraitError, assign_bad)

if __name__ == '__main__':
    unittest.main()



from base64 import b64decode

f = open('./log.txt', 'r')
log = f.read().splitlines()

for i in range(len(log)):
    if log[i] == '----------------------------- Captured stdout call -----------------------------':
        name = log[i + 2]
        print(name)
        base64 = log[i + 3][2:-1]

        with open(name + '.png', 'wb') as f:
            f.write(b64decode(base64))



import k3d
import pytest
from .plot_compare import *
import numpy as np
import skimage.measure


def generate(dim):
    data = np.zeros((dim, dim, dim), dtype=np.uint8)

    x = np.linspace(-0.5, 0.5, dim, dtype=np.float32)
    y = np.linspace(-0.5, 0.5, dim, dtype=np.float32)
    z = np.linspace(-0.5, 0.5, dim, dtype=np.float32)

    x, y, z = np.meshgrid(x, y, z)

    c, s = np.cos(1.5 * x), np.sin(1.5 * x)

    my = y * c - z * s
    mz = y * s + z * c

    my = np.fmod(my + 0.5, 0.333) * 3 - 0.5
    mz = np.fmod(mz + 0.5, 0.333) * 3 - 0.5

    displace = np.sin(60.0 * x) * np.sin(60.0 * my) * np.sin(60.0 * mz) * 0.1

    data = np.sqrt(my ** 2 + mz ** 2) * (2.5 + 0.8 * np.sin(x * 50)) + displace

    return (data < 0.25).astype(np.uint8)


dim = 256
data = generate(dim)

chunk_size = 16
voxelsGroup = []

for z, y, x in zip(*np.where(
        skimage.measure.block_reduce(data, (chunk_size, chunk_size, chunk_size), np.max) > 0)):
    chunk = {
        "voxels": data[z * chunk_size:(z + 1) * chunk_size,
                  y * chunk_size:(y + 1) * chunk_size,
                  x * chunk_size:(x + 1) * chunk_size],
        "coord": np.array([x, y, z]) * chunk_size,
        "multiple": 1
    }

    voxelsGroup.append(chunk)


def test_voxels_group():
    global voxelsGroup, data

    prepare()

    space_size = np.array(data.shape, dtype=np.uint32)[::-1]
    obj = k3d.voxels_group(space_size, voxelsGroup)

    pytest.plot += obj

    compare('voxels_group')


def test_voxels_group_wireframe():
    global voxelsGroup, data

    prepare()

    space_size = np.array(data.shape, dtype=np.uint32)[::-1]
    obj = k3d.voxels_group(space_size, voxelsGroup, wireframe=True, opacity=0.25)

    pytest.plot += obj

    compare('voxels_group_wireframe', camera_factor=0.5)


def test_voxels_group_opacity():
    global voxelsGroup, data

    prepare()

    space_size = np.array(data.shape, dtype=np.uint32)[::-1]
    obj = k3d.voxels_group(space_size, voxelsGroup, opacity=0.5)

    pytest.plot += obj

    compare('voxels_group_opacity')



import k3d
import numpy as np
import pytest
from .plot_compare import *


def test_surface():
    prepare()

    Nx = 240
    Ny = 480
    xmin, xmax = -3, 3
    ymin, ymax = -0, 3

    x = np.linspace(xmin, xmax, Nx)
    y = np.linspace(ymin, ymax, Ny)
    x, y = np.meshgrid(x, y)

    heights = np.sin(x ** 2 + y ** 2).astype(np.float32)

    surface = k3d.surface(heights, xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax)

    pytest.plot += surface

    compare('surface')

    surface.color = 0x00ffff

    compare('surface_dynamic_color')


def test_surface_attribute():
    prepare()

    Nx = 240
    Ny = 480
    xmin, xmax = -3, 3
    ymin, ymax = -0, 3

    x = np.linspace(xmin, xmax, Nx)
    y = np.linspace(ymin, ymax, Ny)
    x, y = np.meshgrid(x, y)

    heights = np.sin(x ** 2 + y ** 2).astype(np.float32)

    surface = k3d.surface(heights, xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax, attribute=heights,
                          transform=k3d.transform(rotation=[np.pi / 4, 1, 0, 0]))

    pytest.plot += surface

    compare('surface_attribute')


def test_surface_attribute_low():
    prepare()

    Nx = 24
    Ny = 48
    xmin, xmax = -3, 3
    ymin, ymax = -0, 3

    x = np.linspace(xmin, xmax, Nx)
    y = np.linspace(ymin, ymax, Ny)
    x, y = np.meshgrid(x, y)

    heights = np.sin(x ** 2 + y ** 2).astype(np.float32)

    surface = k3d.surface(heights, xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax,
                          attribute=heights, transform=k3d.transform(rotation=[np.pi / 4, 1, 0, 0]))

    pytest.plot += surface

    compare('surface_attribute_low')

    surface.flat_shading = False

    compare('surface_attribute_low_dynamic_smooth')

    surface.wireframe = True

    compare('surface_attribute_low_dynamic_wireframe')



import warnings

import ipywidgets as widgets
import numpy as np
from traitlets import (
    Any,
    Bool,
    Bytes,
    Dict,
    Float,
    Int,
    Integer,
    List,
    TraitError,
    Unicode,
    Union,
    validate,
)
from traittypes import Array

from ._version import __version__ as version
from .helpers import (
    array_serialization_wrap,
    callback_serialization_wrap,
    get_bounding_box_point,
    get_bounding_box_points,
    get_bounding_box,
    shape_validation,
    validate_sparse_voxels,
)
from .validation.stl import AsciiStlData, BinaryStlData

EPSILON = np.finfo(np.float32).eps


class TimeSeries(Union):
    def __init__(self, trait):
        if isinstance(trait, list):
            Union.__init__(self, trait + [Dict(t) for t in trait])
        else:
            Union.__init__(self, [trait, Dict(trait)])


class ListOrArray(List):
    _cast_types = (tuple, np.ndarray)

    def __init__(self, *args, **kwargs):
        self._empty_ok = kwargs.pop("empty_ok", False)
        List.__init__(self, *args, **kwargs)

    def validate_elements(self, obj, value):
        if self._empty_ok and len(value) == 0:
            return list(value)
        return super(ListOrArray, self).validate_elements(obj, value)


class VoxelChunk(widgets.Widget):
    """
    Voxel chunk class for selective updating voxels
    """

    _model_name = Unicode("ChunkModel").tag(sync=True)
    _model_module = Unicode("k3d").tag(sync=True)
    _model_module_version = Unicode(version).tag(sync=True)

    id = Int().tag(sync=True)
    voxels = Array(dtype=np.uint8).tag(sync=True, **array_serialization_wrap("voxels"))
    coord = Array(dtype=np.uint32).tag(sync=True, **array_serialization_wrap("coord"))
    multiple = Int().tag(sync=True)
    compression_level = Integer().tag(sync=True)

    def push_data(self, field):
        self.notify_change({"name": field, "type": "change"})

    def __init__(self, **kwargs):
        self.id = id(self)
        super(VoxelChunk, self).__init__(**kwargs)

    def __getitem__(self, name):
        return getattr(self, name)


class Drawable(widgets.Widget):
    """
    Base class for drawable objects and groups.
    """

    _model_name = Unicode("ObjectModel").tag(sync=True)
    _model_module = Unicode("k3d").tag(sync=True)
    _model_module_version = Unicode(version).tag(sync=True)

    id = Integer().tag(sync=True)
    name = Unicode(default_value=None, allow_none=True).tag(sync=True)
    visible = TimeSeries(Bool(True)).tag(sync=True)
    compression_level = Integer().tag(sync=True)

    def __getitem__(self, name):
        return getattr(self, name)

    def __init__(self, **kwargs):
        self.id = id(self)

        super(Drawable, self).__init__(**kwargs)

    def __iter__(self):
        return (self,).__iter__()

    def __add__(self, other):
        return Group(self, other)

    def fetch_data(self, field):
        """Request updating the value of a field modified in browser.

        For data modified in the widget on the browser side, this triggers an asynchronous
        update of the value in the Python kernel.

        Only specific features require this mechanism, e.g. the in-browser editing of voxels.

        Arguments:
            field: `str`.
                The field name."""
        self.send({"msg_type": "fetch", "field": field})

    def push_data(self, field):
        """Request updating the value of a field modified in backend.

        For data modified in the backend side, this triggers an asynchronous
        update of the value in the browser widget.

        Only specific features require this mechanism, e.g. the in-browser editing of voxels.

        Arguments:
            field: `str`.
                The field name."""
        self.notify_change({"name": field, "type": "change"})

    def _ipython_display_(self, **kwargs):
        """Called when `IPython.display.display` is called on the widget."""
        import k3d

        plot = k3d.plot()
        plot += self
        plot.display()


class DrawableWithVoxelCallback(Drawable):
    """
    Base class for drawable with voxels callback handling
    """

    click_callback = None
    hover_callback = None

    def __init__(self, **kwargs):
        super(DrawableWithVoxelCallback, self).__init__(**kwargs)

        self.on_msg(self._handle_custom_msg)

    def _handle_custom_msg(self, content, buffers):
        if content.get("msg_type", "") == "click_callback":
            if self.click_callback is not None:
                self.click_callback(
                    content["coord"]["x"], content["coord"]["y"], content["coord"]["z"]
                )

        if content.get("msg_type", "") == "hover_callback":
            if self.hover_callback is not None:
                self.hover_callback(
                    content["coord"]["x"], content["coord"]["y"], content["coord"]["z"]
                )


class DrawableWithCallback(Drawable):
    """
    Base class for drawable with callback handling
    """

    click_callback = Any(default_value=None, allow_none=True).tag(
        sync=True, **callback_serialization_wrap("click_callback")
    )
    hover_callback = Any(default_value=None, allow_none=True).tag(
        sync=True, **callback_serialization_wrap("hover_callback")
    )

    def __init__(self, **kwargs):
        super(DrawableWithCallback, self).__init__(**kwargs)

        self.on_msg(self._handle_custom_msg)

    def _handle_custom_msg(self, content, buffers):
        if content.get("msg_type", "") == "click_callback":
            if self.click_callback is not None:
                self.click_callback(content)

        if content.get("msg_type", "") == "hover_callback":
            if self.hover_callback is not None:
                self.hover_callback(content)


class Group(Drawable):
    """
    An aggregated group of Drawables, itself a Drawable.

    It can be inserted or removed from a Plot including all members.
    """

    __objs = None

    def __init__(self, *args):
        self.__objs = tuple(
            self.__assert_drawable(drawable)
            for drawables in args
            for drawable in drawables
        )

    def __iter__(self):
        return self.__objs.__iter__()

    def __setattr__(self, key, value):
        """Special method override which allows for setting model matrix for all members of the group."""
        if key == "model_matrix":
            for d in self:
                d.model_matrix = value
        else:
            super(Group, self).__setattr__(key, value)

    @staticmethod
    def __assert_drawable(arg):
        assert isinstance(arg, Drawable)

        return arg


# DRAWABLE OBJECTS


class Line(Drawable):
    """
    A path (polyline) made up of line segments.

    Attributes:
        vertices: `array_like`.
            An array with (x, y, z) coordinates of segment endpoints.
        colors: `array_like`.
            Same-length array of (`int`) packed RGB color of the points (0xff0000 is red, 0xff is blue).
        color: `int`.
            Packed RGB color of the lines (0xff0000 is red, 0xff is blue) when `colors` is empty.
        attribute: `array_like`.
            Array of float attribute for the color mapping, coresponding to each vertex.
        color_map: `list`.
            A list of float quadruplets (attribute value, R, G, B), sorted by attribute value. The first
            quadruplet should have value 0.0, the last 1.0; R, G, B are RGB color components in the range 0.0 to 1.0.
        color_range: `list`.
            A pair [min_value, max_value], which determines the levels of color attribute mapped
            to 0 and 1 in the color map respectively.
        width: `float`.
            The thickness of the lines.
        shader: `str`.
            Display style (name of the shader used) of the lines.
            Legal values are:

            :`simple`: simple lines,

            :`thick`: thick lines,

            :`mesh`: high precision triangle mesh of segments (high quality and GPU load).
        radial_segments: 'int':
            Number of segmented faces around the circumference of the tube.
        model_matrix: `array_like`.
            4x4 model transform matrix.
    """

    type = Unicode(read_only=True).tag(sync=True)

    vertices = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("vertices")
    )
    colors = TimeSeries(Array(dtype=np.uint32)).tag(
        sync=True, **array_serialization_wrap("colors")
    )
    color = TimeSeries(Int(min=0, max=0xFFFFFF)).tag(sync=True)
    width = TimeSeries(Float(min=EPSILON, default_value=0.01)).tag(sync=True)
    attribute = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("attribute")
    )
    color_map = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("color_map")
    )
    color_range = TimeSeries(ListOrArray(minlen=2, maxlen=2, empty_ok=True)).tag(
        sync=True
    )
    shader = TimeSeries(Unicode()).tag(sync=True)
    radial_segments = TimeSeries(Int()).tag(sync=True)
    model_matrix = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("model_matrix")
    )

    def __init__(self, **kwargs):
        super(Line, self).__init__(**kwargs)

        self.set_trait("type", "Line")

    def get_bounding_box(self):
        return get_bounding_box_points(self.vertices, self.model_matrix)

    @validate("colors")
    def _validate_colors(self, proposal):
        if type(proposal["value"]) is dict or type(self.vertices) is dict:
            return proposal["value"]

        required = self.vertices.size // 3  # (x, y, z) triplet per 1 color
        actual = proposal["value"].size
        if actual != 0 and required != actual:
            raise TraitError(
                "colors has wrong size: %s (%s required)" % (actual, required)
            )
        return proposal["value"]


class MarchingCubes(DrawableWithCallback):
    """
    An isosurface in a scalar field obtained through Marching Cubes algorithm.

    The default domain of the scalar field is -0.5 < x, y, z < 0.5.
    If the domain should be different, the bounding box needs to be transformed using the model_matrix.

    Attributes:
        scalar_field: `array_like`.
            A 3D scalar field of values.
        level: `float`.
            Value at the computed isosurface.
        spacings_x: `array_like`.
            A spacings in x axis. Should match to scalar_field shape.
        spacings_y: `array_like`.
            A spacings in y axis. Should match to scalar_field shape.
        spacings_z: `array_like`.
            A spacings in z axis. Should match to scalar_field shape.
        color: `int`.
            Packed RGB color of the isosurface (0xff0000 is red, 0xff is blue).
        wireframe: `bool`.
            Whether mesh should display as wireframe.
        flat_shading: `bool`.
            Whether mesh should display with flat shading.
        opacity: `float`.
            Opacity of mesh.
        model_matrix: `array_like`.
            4x4 model transform matrix.
    """

    type = Unicode(read_only=True).tag(sync=True)
    scalar_field = Array(dtype=np.float32).tag(
        sync=True, **array_serialization_wrap("scalar_field")
    )
    spacings_x = Array(dtype=np.float32).tag(
        sync=True, **array_serialization_wrap("spacings_x")
    )
    spacings_y = Array(dtype=np.float32).tag(
        sync=True, **array_serialization_wrap("spacings_y")
    )
    spacings_z = Array(dtype=np.float32).tag(
        sync=True, **array_serialization_wrap("spacings_z")
    )
    level = Float().tag(sync=True)
    color = Int(min=0, max=0xFFFFFF).tag(sync=True)
    wireframe = Bool().tag(sync=True)
    flat_shading = Bool().tag(sync=True)
    opacity = TimeSeries(Float(min=0.0, max=1.0, default_value=1.0)).tag(sync=True)
    model_matrix = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("model_matrix")
    )

    def get_bounding_box(self):
        return get_bounding_box(self.model_matrix)

    def __init__(self, **kwargs):
        super(MarchingCubes, self).__init__(**kwargs)

        self.set_trait("type", "MarchingCubes")


class Mesh(DrawableWithCallback):
    """
    A 3D triangles mesh.

    Attributes:
        vertices: `array_like`.
            Array of triangle vertices: float (x, y, z) coordinate triplets.
        indices: `array_like`.
            Array of vertex indices: int triplets of indices from vertices array.
        color: `int`.
            Packed RGB color of the mesh (0xff0000 is red, 0xff is blue) when not using color maps.
        colors: `array_like`.
            Same-length array of (`int`) packed RGB color of the points (0xff0000 is red, 0xff is blue).
        attribute: `array_like`.
            Array of float attribute for the color mapping, coresponding to each vertex.
        triangles_attribute: `array_like`.
            Array of float attribute for the color mapping, coresponding to each triangle.
        color_map: `list`.
            A list of float quadruplets (attribute value, R, G, B), sorted by attribute value. The first
            quadruplet should have value 0.0, the last 1.0; R, G, B are RGB color components in the range 0.0 to 1.0.
        color_range: `list`.
            A pair [min_value, max_value], which determines the levels of color attribute mapped
            to 0 and 1 in the color map respectively.
        wireframe: `bool`.
            Whether mesh should display as wireframe.
        flat_shading: `bool`.
            Whether mesh should display with flat shading.
        opacity: `float`.
            Opacity of mesh.
        volume: `array_like`.
            3D array of `float`
        volume_bounds: `array_like`.
            6-element tuple specifying the bounds of the volume data (x0, x1, y0, y1, z0, z1)
        texture: `bytes`.
            Image data in a specific format.
        texture_file_format: `str`.
            Format of the data, it should be the second part of MIME format of type 'image/',
            for example 'jpeg', 'png', 'gif', 'tiff'.
        uvs: `array_like`.
            Array of float uvs for the texturing, coresponding to each vertex.
        model_matrix: `array_like`.
            4x4 model transform matrix.
    """

    type = Unicode(read_only=True).tag(sync=True)
    vertices = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("vertices")
    )
    indices = TimeSeries(Array(dtype=np.uint32)).tag(
        sync=True, **array_serialization_wrap("indices")
    )
    color = TimeSeries(Int(min=0, max=0xFFFFFF)).tag(sync=True)
    colors = TimeSeries(Array(dtype=np.uint32)).tag(
        sync=True, **array_serialization_wrap("colors")
    )
    attribute = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("attribute")
    )
    triangles_attribute = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("triangles_attribute")
    )
    color_map = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("color_map")
    )
    color_range = TimeSeries(ListOrArray(minlen=2, maxlen=2, empty_ok=True)).tag(
        sync=True
    )
    wireframe = TimeSeries(Bool()).tag(sync=True)
    flat_shading = TimeSeries(Bool()).tag(sync=True)
    side = TimeSeries(Unicode()).tag(sync=True)
    opacity = TimeSeries(Float(min=0.0, max=1.0, default_value=1.0)).tag(sync=True)
    volume = TimeSeries(Array()).tag(sync=True, **array_serialization_wrap("volume"))
    volume_bounds = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("volume_bounds")
    )
    texture = Bytes(allow_none=True).tag(sync=True, **array_serialization_wrap("texture"))
    texture_file_format = Unicode(allow_none=True).tag(sync=True)
    uvs = TimeSeries(Array()).tag(sync=True, **array_serialization_wrap("uvs"))
    opacity_function = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("opacity_function")
    )
    model_matrix = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("model_matrix")
    )

    def __init__(self, **kwargs):
        super(Mesh, self).__init__(**kwargs)

        self.set_trait("type", "Mesh")

    @validate("colors")
    def _validate_colors(self, proposal):
        if type(proposal["value"]) is dict or type(self.vertices) is dict:
            return proposal["value"]

        required = self.vertices.size // 3  # (x, y, z) triplet per 1 color
        actual = proposal["value"].size
        if actual != 0 and required != actual:
            raise TraitError(
                "colors has wrong size: %s (%s required)" % (actual, required)
            )
        return proposal["value"]

    @validate("volume")
    def _validate_volume(self, proposal):
        if type(proposal["value"]) is dict:
            return proposal["value"]

        if type(proposal["value"]) is np.ndarray and proposal[
            "value"
        ].dtype is np.dtype(object):
            return proposal["value"].tolist()

        if proposal["value"].shape == (0,):
            return np.array(proposal["value"], dtype=np.float32)

        required = [np.float16, np.float32]
        actual = proposal["value"].dtype

        if actual not in required:
            warnings.warn("wrong dtype: %s (%s required)" % (actual, required))

            return proposal["value"].astype(np.float32)

        return proposal["value"]

    def get_bounding_box(self):
        return get_bounding_box_points(self.vertices, self.model_matrix)


class Points(Drawable):
    """
    A point cloud.

    Attributes:
        positions: `array_like`.
            Array with (x, y, z) coordinates of the points.
        colors: `array_like`.
            Same-length array of (`int`) packed RGB color of the points (0xff0000 is red, 0xff is blue).
        color: `int`.
            Packed RGB color of the points (0xff0000 is red, 0xff is blue) when `colors` is empty.
        point_size: `float`.
            Diameter of the balls representing the points in 3D space.
        point_sizes: `array_like`.
            Same-length array of `float` sizes of the points.
        shader: `str`.
            Display style (name of the shader used) of the points.
            Legal values are:

            :`flat`: simple circles with uniform color,

            :`dot`: simple dot with uniform color,

            :`3d`: little 3D balls,

            :`3dSpecular`: little 3D balls with specular lightning,

            :`mesh`: high precision triangle mesh of a ball (high quality and GPU load).
        mesh_detail: `int`.
            Default is 2. Setting this to a value greater than 0 adds more vertices making it no longer an
            icosahedron. When detail is greater than 1, it's effectively a sphere. Only valid if shader='mesh'
        attribute: `array_like`.
            Array of float attribute for the color mapping, coresponding to each point.
        color_map: `list`.
            A list of float quadruplets (attribute value, R, G, B), sorted by attribute value. The first
            quadruplet should have value 0.0, the last 1.0; R, G, B are RGB color components in the range 0.0 to 1.0.
        color_range: `list`.
            A pair [min_value, max_value], which determines the levels of color attribute mapped
            to 0 and 1 in the color map respectively.
        opacity_function: `array`.
            A list of float tuples (attribute value, opacity), sorted by attribute value. The first
            tuples should have value 0.0, the last 1.0; opacity is in the range 0.0 to 1.0.
        model_matrix: `array_like`.
            4x4 model transform matrix.
    """

    type = Unicode(read_only=True).tag(sync=True)
    positions = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("positions")
    )
    colors = TimeSeries(Array(dtype=np.uint32)).tag(
        sync=True, **array_serialization_wrap("colors")
    )
    color = TimeSeries(Int(min=0, max=0xFFFFFF)).tag(sync=True)
    point_size = TimeSeries(Float(min=EPSILON, default_value=1.0)).tag(sync=True)
    point_sizes = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("point_sizes")
    )
    opacity = TimeSeries(Float(min=0.0, max=1.0, default_value=1.0)).tag(sync=True)
    opacities = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("opacities")
    )
    shader = TimeSeries(Unicode()).tag(sync=True)
    mesh_detail = TimeSeries(Int(min=0, max=8)).tag(sync=True)
    attribute = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("attribute")
    )
    color_map = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("color_map")
    )
    color_range = TimeSeries(ListOrArray(minlen=2, maxlen=2, empty_ok=True)).tag(
        sync=True
    )
    opacity_function = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("opacity_function")
    )
    model_matrix = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("model_matrix")
    )

    def __init__(self, **kwargs):
        super(Points, self).__init__(**kwargs)

        self.set_trait("type", "Points")

    @validate("colors")
    def _validate_colors(self, proposal):
        if type(proposal["value"]) is dict or type(self.positions) is dict:
            return proposal["value"]

        required = self.positions.size // 3  # (x, y, z) triplet per 1 color
        actual = proposal["value"].size
        if actual != 0 and required != actual:
            raise TraitError(
                "colors has wrong size: %s (%s required)" % (actual, required)
            )
        return proposal["value"]

    def get_bounding_box(self):
        return get_bounding_box_points(self.positions, self.model_matrix)


class STL(Drawable):
    """
    A STereoLitograpy 3D geometry.

    STL is a popular format introduced for 3D printing. There are two sub-formats - ASCII and binary.

    Attributes:
        text: `str`.
            STL data in text format (ASCII STL).
        binary: `bytes`.
            STL data in binary format (Binary STL).
            The `text` attribute should be set to None when using Binary STL.
        color: `int`.
            Packed RGB color of the resulting mesh (0xff0000 is red, 0xff is blue).
        model_matrix: `array_like`.
            4x4 model transform matrix.
        wireframe: `bool`.
            Whether mesh should display as wireframe.
        flat_shading: `bool`.
            Whether mesh should display with flat shading.
    """

    type = Unicode(read_only=True).tag(sync=True)
    text = AsciiStlData(allow_none=True, default_value=None).tag(sync=True)
    binary = BinaryStlData(allow_none=True,
                           default_value=None).tag(sync=True, **array_serialization_wrap("binary"))
    color = Int(min=0, max=0xFFFFFF).tag(sync=True)
    wireframe = Bool().tag(sync=True)
    flat_shading = Bool().tag(sync=True)
    model_matrix = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("model_matrix")
    )

    def __init__(self, **kwargs):
        super(STL, self).__init__(**kwargs)

        self.set_trait("type", "STL")

    def get_bounding_box(self):
        warnings.warn("STL bounding box is still not supported")
        return [-1, 1, -1, 1, -1, 1]


class Surface(DrawableWithCallback):
    """
    Surface plot of a 2D function z = f(x, y).

    The default domain of the scalar field is -0.5 < x, y < 0.5.
    If the domain should be different, the bounding box needs to be transformed using the model_matrix.

    Attributes:
        heights: `array_like`.
            2D scalar field of Z values.
        color: `int`.
            Packed RGB color of the resulting mesh (0xff0000 is red, 0xff is blue).
        wireframe: `bool`.
            Whether mesh should display as wireframe.
        flat_shading: `bool`.
            Whether mesh should display with flat shading.
        attribute: `array_like`.
            Array of float attribute for the color mapping, coresponding to each vertex.
        color_map: `list`.
            A list of float quadruplets (attribute value, R, G, B), sorted by attribute value. The first
            quadruplet should have value 0.0, the last 1.0; R, G, B are RGB color components in the range 0.0 to 1.0.
        color_range: `list`.
            A pair [min_value, max_value], which determines the levels of color attribute mapped
            to 0 and 1 in the color map respectively.
        model_matrix: `array_like`.
            4x4 model transform matrix.
    """

    type = Unicode(read_only=True).tag(sync=True)
    heights = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("heights")
    )
    color = Int(min=0, max=0xFFFFFF).tag(sync=True)
    wireframe = Bool().tag(sync=True)
    flat_shading = Bool().tag(sync=True)
    attribute = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("attribute")
    )
    color_map = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("color_map")
    )
    color_range = TimeSeries(ListOrArray(minlen=2, maxlen=2, empty_ok=True)).tag(
        sync=True
    )
    model_matrix = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("model_matrix")
    )

    def __init__(self, **kwargs):
        super(Surface, self).__init__(**kwargs)

        self.set_trait("type", "Surface")

    def get_bounding_box(self):
        return get_bounding_box(self.model_matrix)


class Text(Drawable):
    """
    Text rendered using KaTeX with a 3D position.

    Attributes:
        text: `str`.
            Content of the text.
        position: `list`.
            Coordinates (x, y, z) of the text's position.
        color: `int`.
            Packed RGB color of the text (0xff0000 is red, 0xff is blue).
        is_html: `Boolean`.
            Whether text should be interpreted as HTML insted of KaTeX.
        on_top: `Boolean`.
            Render order with 3d object
        reference_point: `str`.
            Two-letter string representing the text's alignment.

            First letter: 'l', 'c' or 'r': left, center or right

            Second letter: 't', 'c' or 'b': top, center or bottom.
        size: `float`.
            Font size in 'em' HTML units.
        label_box: `Boolean`.
            Label background box.
    """

    type = Unicode(read_only=True).tag(sync=True)
    text = TimeSeries(Unicode()).tag(sync=True)
    position = TimeSeries(ListOrArray(minlen=3, maxlen=3)).tag(sync=True)
    is_html = Bool(False).tag(sync=True)
    color = Int(min=0, max=0xFFFFFF).tag(sync=True)
    reference_point = Unicode().tag(sync=True)
    size = TimeSeries(Float(min=EPSILON, default_value=1.0)).tag(sync=True)
    on_top = Bool().tag(sync=True)
    label_box = Bool().tag(sync=True)

    def __init__(self, **kwargs):
        super(Text, self).__init__(**kwargs)

        self.set_trait("type", "Text")

    def get_bounding_box(self):
        return get_bounding_box_point(self.position)


class Text2d(Drawable):
    """
    Text rendered using KaTeX with a fixed 2D position, independent of camera settings.

    Attributes:
        text: `str`.
            Content of the text.
        position: `list`.
            Ratios (r_x, r_y) of the text's position in range (0, 1) - relative to canvas size.
        color: `int`.
            Packed RGB color of the text (0xff0000 is red, 0xff is blue).
        is_html: `Boolean`.
            Whether text should be interpreted as HTML insted of KaTeX.
        reference_point: `str`.
            Two-letter string representing the text's alignment.

            First letter: 'l', 'c' or 'r': left, center or right

            Second letter: 't', 'c' or 'b': top, center or bottom.
        size: `float`.
            Font size in 'em' HTML units.
        label_box: `Boolean`.
            Label background box.
    """

    type = Unicode(read_only=True).tag(sync=True)
    color = Int(min=0, max=0xFFFFFF).tag(sync=True)
    size = TimeSeries(Float(min=EPSILON, default_value=1.0)).tag(sync=True)
    is_html = Bool(False).tag(sync=True)
    reference_point = Unicode().tag(sync=True)
    position = TimeSeries(ListOrArray(minlen=2, maxlen=2)).tag(sync=True)
    text = TimeSeries(Unicode()).tag(sync=True)
    label_box = Bool().tag(sync=True)

    def __init__(self, **kwargs):
        super(Text2d, self).__init__(**kwargs)

        self.set_trait("type", "Text2d")

    def get_bounding_box(self):
        return get_bounding_box_point(self.position)


class Label(Drawable):
    """
    Label rendered using KaTeX with a 3D position.

    Attributes:
        text: `str`.
            Content of the text.
        position: `list`.
            Coordinates (x, y, z) of the text's position.
        color: `int`.
            Packed RGB color of the text (0xff0000 is red, 0xff is blue).
        on_top: `Boolean`.
            Render order with 3d object
        label_box: `Boolean`.
            Label background box.
        mode: `str`.
            Label node. Can be 'dynamic', 'local' or 'side'.
        is_html: `Boolean`.
            Whether text should be interpreted as HTML insted of KaTeX.
        max_length: `float`.
            Maximum length of line in % of half screen size.
        size: `float`.
            Font size in 'em' HTML units.
    """

    type = Unicode(read_only=True).tag(sync=True)
    mode = Unicode().tag(sync=True)
    text = TimeSeries(Unicode()).tag(sync=True)
    is_html = Bool(False).tag(sync=True)
    position = TimeSeries(ListOrArray(minlen=3, maxlen=3)).tag(sync=True)
    color = Int(min=0, max=0xFFFFFF).tag(sync=True)
    max_length = Float(min=0, max=1.0).tag(sync=True)
    size = TimeSeries(Float(min=EPSILON, default_value=1.0)).tag(sync=True)
    on_top = Bool().tag(sync=True)
    label_box = Bool().tag(sync=True)

    def __init__(self, **kwargs):
        super(Label, self).__init__(**kwargs)

        self.set_trait("type", "Label")

    def get_bounding_box(self):
        return get_bounding_box_point(self.position)


class Texture(DrawableWithCallback):
    """
    A 2D image displayed as a texture.

    By default, the texture image is mapped into the square: -0.5 < x, y < 0.5, z = 1.
    If the size (scale, aspect ratio) or position should be different then the texture should be transformed
    using the model_matrix.

    Attributes:
        binary: `bytes`.
            Image data in a specific format.
        file_format: `str`.
            Format of the data, it should be the second part of MIME format of type 'image/',
            for example 'jpeg', 'png', 'gif', 'tiff'.
        attribute: `array_like`.
            Array of float attribute for the color mapping, coresponding to each pixels.
        color_map: `list`.
            A list of float quadruplets (attribute value, R, G, B), sorted by attribute value. The first
            quadruplet should have value 0.0, the last 1.0; R, G, B are RGB color components in the range 0.0 to 1.0.
        opacity_function: `array`.
            A list of float tuples (attribute value, opacity), sorted by attribute value. The first
            tuples should have value 0.0, the last 1.0; opacity is in the range 0.0 to 1.0.
        color_range: `list`.
            A pair [min_value, max_value], which determines the levels of color attribute mapped
            to 0 and 1 in the color map respectively.
        interpolation: `bool`.
            Whether data should be interpolatedor not.
        puv: `list`.
            A list of float triplets (x,y,z). The first triplet mean a position of left-bottom corner of texture.
            Second and third triplets means a base of coordinate system for texture.
        model_matrix: `array_like`.
            4x4 model transform matrix.
    """

    type = Unicode(read_only=True).tag(sync=True)
    binary = Bytes(allow_none=True).tag(sync=True, **array_serialization_wrap("binary"))
    file_format = Unicode(allow_none=True).tag(sync=True)
    attribute = Array().tag(sync=True, **array_serialization_wrap("attribute"))
    puv = Array(dtype=np.float32).tag(sync=True, **array_serialization_wrap("puv"))
    color_map = Array(dtype=np.float32).tag(
        sync=True, **array_serialization_wrap("color_map")
    )
    color_range = ListOrArray(minlen=2, maxlen=2, empty_ok=True).tag(sync=True)
    interpolation = TimeSeries(Bool()).tag(sync=True)
    opacity_function = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("opacity_function")
    )
    model_matrix = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("model_matrix")
    )

    def __init__(self, **kwargs):
        super(Texture, self).__init__(**kwargs)

        self.set_trait("type", "Texture")

    def get_bounding_box(self):
        return get_bounding_box(self.model_matrix)


class TextureText(Drawable):
    """
    A text in the 3D space rendered using a texture.

    Compared to Text and Text2d this drawable has less features (no KaTeX support), but the labels are located
    in the GPU memory, and not the browser's DOM tree. This has performance consequences, and may be preferable when
    many simple labels need to be displayed.

    Attributes:
        text: `str`.
            Content of the text.
        position: `list`.
            Coordinates (x, y, z) of the text's position.
        color: `int`.
            Packed RGB color of the text (0xff0000 is red, 0xff is blue).
        size: `float`.
            Size of the texture sprite containing the text.
        font_face: `str`.
            Name of the font to use for rendering the text.
        font_weight: `int`.
            Thickness of the characters in HTML-like units from the range (100, 900), where
            400 is normal and 600 is bold font.
        font_size: `int`.
            The font size inside the sprite texture in px units. This does not affect the size of the
            text in the scene, only the accuracy and raster size of the texture.
    """

    type = Unicode(read_only=True).tag(sync=True)
    text = TimeSeries(Unicode()).tag(sync=True)
    position = TimeSeries(ListOrArray(minlen=3, maxlen=3)).tag(sync=True)
    color = TimeSeries(Int(min=0, max=0xFFFFFF)).tag(sync=True)
    size = TimeSeries(Float(min=EPSILON, default_value=1.0)).tag(sync=True)
    font_face = Unicode().tag(sync=True)
    font_weight = Int().tag(sync=True)
    font_size = Int().tag(sync=True)

    def __init__(self, **kwargs):
        super(TextureText, self).__init__(**kwargs)

        self.set_trait("type", "TextureText")

    def get_bounding_box(self):
        return get_bounding_box_point(self.position)


class VectorField(Drawable):
    """
    A dense 3D or 2D vector field.

    By default, the origins of the vectors are assumed to be a grid inscribed in the -0.5 < x, y, z < 0.5 cube
    or -0.5 < x, y < 0.5 square, regardless of the passed vector field shape (aspect ratio etc.).
    Different grid size, shape and rotation can be obtained using the model_matrix.

    The color of the vectors is a gradient from origin_color to head_color. Heads, when used, have uniform head_color.

    For sparse (i.e. not forming a grid) 3D vectors, use the `Vectors` drawable.

    Attributes:
        vectors: `array_like`.
            Vector field of shape (L, H, W, 3) for 3D fields or (H, W, 2) for 2D fields.
        colors: `array_like`.
            Twice the length of vectors array of int: packed RGB colors
            (0xff0000 is red, 0xff is blue).
            The array has consecutive pairs (origin_color, head_color) for vectors in row-major order.
        origin_color: `int`.
            Packed RGB color of the origins (0xff0000 is red, 0xff is blue) when `colors` is empty.
        head_color: `int`.
            Packed RGB color of the vector heads (0xff0000 is red, 0xff is blue) when `colors` is empty.
        use_head: `bool`.
            Whether vectors should display an arrow head.
        head_size: `float`.
            The size of the arrow heads.
        scale: `float`.
            Scale factor for the vector lengths, for artificially scaling the vectors in place.
        line_width: `float`.
            Width of the vector segments.
        model_matrix: `array_like`.
            4x4 model transform matrix.
    """

    type = Unicode(read_only=True).tag(sync=True)
    vectors = Array(dtype=np.float32).tag(
        sync=True, **array_serialization_wrap("vectors")
    )
    colors = Array(dtype=np.uint32).tag(sync=True, **array_serialization_wrap("colors"))
    origin_color = Int(min=0, max=0xFFFFFF).tag(sync=True)
    head_color = Int(min=0, max=0xFFFFFF).tag(sync=True)
    use_head = Bool().tag(sync=True)
    head_size = Float(min=EPSILON, default_value=1.0).tag(sync=True)
    scale = Float().tag(sync=True)
    line_width = Float(min=EPSILON, default_value=0.01).tag(sync=True)
    model_matrix = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("model_matrix")
    )

    def __init__(self, **kwargs):
        super(VectorField, self).__init__(**kwargs)

        self.set_trait("type", "VectorField")

    @validate("vectors")
    def _validate_vectors(self, proposal):
        shape = proposal["value"].shape
        if len(shape) not in (3, 4) or len(shape) != shape[-1] + 1:
            raise TraitError(
                "Vector field has invalid shape: {}, "
                "expected (L, H, W, 3) for a 3D or (H, W, 2) for a 2D field".format(
                    shape
                )
            )
        return np.array(proposal["value"], np.float32)

    def get_bounding_box(self):
        return get_bounding_box(self.model_matrix)


class Vectors(Drawable):
    """
    3D vectors.

    The color of the vectors is a gradient from origin_color to head_color. Heads, when used, have uniform head_color.

    For dense (i.e. forming a grid) 3D or 2D vectors, use the `VectorField` drawable.

    Attributes:
        vectors: `array_like`.
            The vectors as (dx, dy, dz) float triples.
        origins: `array_like`.
            Same-size array of (x, y, z) coordinates of vector origins.
        colors: `array_like`.
            Twice the length of vectors array of int: packed RGB colors
            (0xff0000 is red, 0xff is blue).
            The array has consecutive pairs (origin_color, head_color) for vectors in row-major order.
        origin_color: `int`.
            Packed RGB color of the origins (0xff0000 is red, 0xff is blue), default: same as color.
        head_color: `int`.
            Packed RGB color of the vector heads (0xff0000 is red, 0xff is blue), default: same as color.
        use_head: `bool`.
            Whether vectors should display an arrow head.
        head_size: `float`.
            The size of the arrow heads.
        labels: `list` of `str`.
            Captions to display next to the vectors.
        label_size: `float`.
            Label font size in 'em' HTML units.
        line_width: `float`.
            Width of the vector segments.
        model_matrix: `array_like`.
            4x4 model transform matrix.
    """

    type = Unicode(read_only=True).tag(sync=True)
    origins = Array(dtype=np.float32).tag(
        sync=True, **array_serialization_wrap("origins")
    )
    vectors = Array(dtype=np.float32).tag(
        sync=True, **array_serialization_wrap("vectors")
    )
    colors = Array(dtype=np.uint32).tag(sync=True, **array_serialization_wrap("colors"))
    origin_color = Int(min=0, max=0xFFFFFF).tag(sync=True)
    head_color = Int(min=0, max=0xFFFFFF).tag(sync=True)
    use_head = Bool().tag(sync=True)
    head_size = Float(min=EPSILON, default_value=1.0).tag(sync=True)
    labels = List().tag(sync=True)
    label_size = Float(min=EPSILON, default_value=1.0).tag(sync=True)
    line_width = Float(min=EPSILON, default_value=0.01).tag(sync=True)
    model_matrix = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("model_matrix")
    )

    def __init__(self, **kwargs):
        super(Vectors, self).__init__(**kwargs)

        self.set_trait("type", "Vectors")

    def get_bounding_box(self):
        return get_bounding_box_points(
            np.stack([self.origins, self.vectors]), self.model_matrix
        )


class Volume(Drawable):
    """
    3D volumetric data.

    By default, the volume are a grid inscribed in the -0.5 < x, y, z < 0.5 cube
    regardless of the passed voxel array shape (aspect ratio etc.).

    Attributes:
        volume: `array_like`.
            3D array of `float`.
        color_map: `array_like`.
            A list of float quadruplets (attribute value, R, G, B), sorted by attribute value. The first
            quadruplet should have value 0.0, the last 1.0; R, G, B are RGB color components in the range 0.0 to 1.0.
        opacity_function: `array`.
            A list of float tuples (attribute value, opacity), sorted by attribute value. The first
            typles should have value 0.0, the last 1.0; opacity is in the range 0.0 to 1.0.
        color_range: `list`.
            A pair [min_value, max_value], which determines the levels of color attribute mapped
            to 0 and 1 in the color map respectively.
        samples: `float`.
            Number of iteration per 1 unit of space.
        alpha_coef: `float`.
            Alpha multiplier.
        shadow: `str`.
            Type of shadow on volume.

            Legal values are:

            :`off`: shadow disabled,

            :`on_demand`: update shadow map on demand ( self.shadow_map_update() ),

            :`dynamic`: update shadow map automaticaly every shadow_delay.
        shadow_delay: `float`.
            Minimum number of miliseconds between shadow map updates.
        shadow_res: `int`.
            Resolution of shadow map.
        interpolation: `bool`.
            Whether volume raycasting should interpolate data or not.
        model_matrix: `array_like`.
            4x4 model transform matrix.
    """

    type = Unicode(read_only=True).tag(sync=True)
    volume = TimeSeries(Array()).tag(sync=True, **array_serialization_wrap("volume"))
    color_map = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("color_map")
    )
    opacity_function = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("opacity_function")
    )
    color_range = TimeSeries(ListOrArray(minlen=2, maxlen=2, empty_ok=True)).tag(
        sync=True
    )
    samples = TimeSeries(Float()).tag(sync=True)
    alpha_coef = TimeSeries(Float()).tag(sync=True)
    gradient_step = TimeSeries(Float()).tag(sync=True)
    shadow = TimeSeries(Unicode()).tag(sync=True)
    shadow_res = TimeSeries(Int(min=31, max=513, default_value=128)).tag(sync=True)
    shadow_delay = TimeSeries(Float()).tag(sync=True)
    ray_samples_count = TimeSeries(Int(min=1, max=128, default_value=16)).tag(sync=True)
    focal_length = TimeSeries(Float()).tag(sync=True)
    focal_plane = TimeSeries(Float()).tag(sync=True)
    interpolation = TimeSeries(Bool()).tag(sync=True)
    model_matrix = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("model_matrix")
    )

    def __init__(self, **kwargs):
        super(Volume, self).__init__(**kwargs)

        self.set_trait("type", "Volume")

    @validate("volume")
    def _validate_volume(self, proposal):
        if type(proposal["value"]) is dict:
            return proposal["value"]

        if type(proposal["value"]) is np.ndarray and proposal[
            "value"
        ].dtype is np.dtype(object):
            return proposal["value"].tolist()

        required = [np.float16, np.float32]
        actual = proposal["value"].dtype

        if actual not in required:
            warnings.warn("wrong dtype: %s (%s required)" % (actual, required))

            return proposal["value"].astype(np.float32)

        return proposal["value"]

    def shadow_map_update(self, direction=None):
        """Request updating the shadow map in browser."""

        self.send({"msg_type": "shadow_map_update", "direction": direction})

    def get_bounding_box(self):
        return get_bounding_box(self.model_matrix)


class MIP(Drawable):
    """
    3D volumetric data.

    By default, the volume are a grid inscribed in the -0.5 < x, y, z < 0.5 cube
    regardless of the passed voxel array shape (aspect ratio etc.).

    Attributes:
        volume: `array_like`.
            3D array of `float`.
        color_map: `array_like`.
            A list of float quadruplets (attribute value, R, G, B), sorted by attribute value. The first
            quadruplet should have value 0.0, the last 1.0; R, G, B are RGB color components in the range 0.0 to 1.0.
        opacity_function: `array`.
            A list of float tuples (attribute value, opacity), sorted by attribute value. The first
            typles should have value 0.0, the last 1.0; opacity is in the range 0.0 to 1.0.
        color_range: `list`.
            A pair [min_value, max_value], which determines the levels of color attribute mapped
            to 0 and 1 in the color map respectively.
        samples: `float`.
            Number of iteration per 1 unit of space.
        gradient_step: `float`
            Gradient light step.
        model_matrix: `array_like`.
            4x4 model transform matrix.
    """

    type = Unicode(read_only=True).tag(sync=True)
    volume = TimeSeries(Array()).tag(sync=True, **array_serialization_wrap("volume"))
    color_map = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("color_map")
    )
    opacity_function = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("opacity_function")
    )
    color_range = TimeSeries(ListOrArray(minlen=2, maxlen=2, empty_ok=True)).tag(
        sync=True
    )
    gradient_step = TimeSeries(Float()).tag(sync=True)
    samples = TimeSeries(Float()).tag(sync=True)
    model_matrix = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("model_matrix")
    )

    def __init__(self, **kwargs):
        super(MIP, self).__init__(**kwargs)

        self.set_trait("type", "MIP")

    @validate("volume")
    def _validate_volume(self, proposal):
        if type(proposal["value"]) is dict:
            return proposal["value"]

        if type(proposal["value"]) is np.ndarray and proposal[
            "value"
        ].dtype is np.dtype(object):
            return proposal["value"].tolist()

        required = [np.float16, np.float32]
        actual = proposal["value"].dtype

        if actual not in required:
            warnings.warn("wrong dtype: %s (%s required)" % (actual, required))

            return proposal["value"].astype(np.float32)

        return proposal["value"]

    def get_bounding_box(self):
        return get_bounding_box(self.model_matrix)


class Voxels(DrawableWithVoxelCallback):
    """
    3D volumetric data.

    Different grid size, shape and rotation can be obtained using model_matrix.

    Attributes:
        voxels: `array_like`.
            3D array of `int` in range (0, 255).
            0 means empty voxel, 1 and above refer to consecutive color_map entries.
        color_map: `array_like`.
            Flat array of `int` packed RGB colors (0xff0000 is red, 0xff is blue).

            The color defined at index i is for voxel value (i+1), e.g.:

           | color_map = [0xff, 0x00ff]
           | voxels =
           | [
           | 0, # empty voxel
           | 1, # blue voxel
           | 2  # red voxel
           | ]

        model_matrix: `array_like`.
            4x4 model transform matrix.
        wireframe: `bool`.
            Whether mesh should display as wireframe.
        opacity: `float`.
            Opacity of voxels.
        outlines: `bool`.
            Whether mesh should display with outlines.
        outlines_color: `int`.
            Packed RGB color of the resulting outlines (0xff0000 is red, 0xff is blue)
    """

    type = Unicode(read_only=True).tag(sync=True)
    voxels = Array(dtype=np.uint8).tag(sync=True, **array_serialization_wrap("voxels"))
    color_map = Array(dtype=np.uint32).tag(
        sync=True, **array_serialization_wrap("voxels")
    )
    wireframe = Bool().tag(sync=True)
    outlines = Bool().tag(sync=True)
    outlines_color = Int(min=0, max=0xFFFFFF).tag(sync=True)
    opacity = TimeSeries(Float(min=0.0, max=1.0, default_value=1.0)).tag(sync=True)
    model_matrix = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("model_matrix")
    )

    def __init__(self, **kwargs):
        super(Voxels, self).__init__(**kwargs)

        self.set_trait("type", "Voxels")

    def get_bounding_box(self):
        return get_bounding_box(self.model_matrix)


class SparseVoxels(DrawableWithVoxelCallback):
    """
    3D volumetric data.

    By default, the voxels are a grid inscribed in the -0.5 < x, y, z < 0.5 cube
    regardless of the passed voxel array shape (aspect ratio etc.).
    Different grid size, shape and rotation can be obtained using the model_matrix.

    Attributes:
        sparse_voxels: `array_like`.
            2D array of `coords` in format [[x,y,z,v],[x,y,z,v]].
            v = 0 means empty voxel, 1 and above refer to consecutive color_map entries.
        space_size: `array_like`.
            Width, Height, Length of space
        color_map: `array_like`.
            Flat array of `int` packed RGB colors (0xff0000 is red, 0xff is blue).
        model_matrix: `array_like`.
            4x4 model transform matrix.
        wireframe: `bool`.
            Whether mesh should display as wireframe.
        opacity: `float`.
            Opacity of voxels.
        outlines: `bool`.
            Whether mesh should display with outlines.
        outlines_color: `int`.
            Packed RGB color of the resulting outlines (0xff0000 is red, 0xff is blue)
    """

    type = Unicode(read_only=True).tag(sync=True)
    sparse_voxels = (
        Array(dtype=np.uint16)
            .tag(sync=True, **array_serialization_wrap("sparse_voxels"))
            .valid(validate_sparse_voxels)
    )
    space_size = (
        Array(dtype=np.uint32)
            .tag(sync=True, **array_serialization_wrap("space_size"))
            .valid(shape_validation(3))
    )
    color_map = Array(dtype=np.uint32).tag(
        sync=True, **array_serialization_wrap("color_map")
    )
    wireframe = Bool().tag(sync=True)
    outlines = Bool().tag(sync=True)
    outlines_color = Int(min=0, max=0xFFFFFF).tag(sync=True)
    opacity = TimeSeries(Float(min=0.0, max=1.0, default_value=1.0)).tag(sync=True)
    model_matrix = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("model_matrix")
    )

    def __init__(self, **kwargs):
        super(SparseVoxels, self).__init__(**kwargs)

        self.set_trait("type", "SparseVoxels")

    def get_bounding_box(self):
        return get_bounding_box(self.model_matrix)


class VoxelsGroup(DrawableWithVoxelCallback):
    """
    3D volumetric data.

    By default, the voxels are a grid inscribed in the -0.5 < x, y, z < 0.5 cube
    regardless of the passed voxel array shape (aspect ratio etc.).
    Different grid size, shape and rotation can be obtained using the model_matrix.

    Attributes:
        voxels_group: `array_like`.
            List of `chunks` in format {voxels: np.array, coord: [x,y,z], multiple: number}.
        space_size: `array_like`.
            Width, Height, Length of space
        color_map: `array_like`.
            Flat array of `int` packed RGB colors (0xff0000 is red, 0xff is blue).
        model_matrix: `array_like`.
            4x4 model transform matrix.
        wireframe: `bool`.
            Whether mesh should display as wireframe.
        opacity: `float`.
            Opacity of voxels.
        outlines: `bool`.
            Whether mesh should display with outlines.
        outlines_color: `int`.
            Packed RGB color of the resulting outlines (0xff0000 is red, 0xff is blue)
    """

    type = Unicode(read_only=True).tag(sync=True)

    _hold_remeshing = Bool(default_value=False).tag(sync=True)

    voxels_group = List().tag(sync=True, **array_serialization_wrap("voxels_group"))
    chunks_ids = List().tag(sync=True)

    space_size = Array(dtype=np.uint32).tag(
        sync=True, **array_serialization_wrap("space_size")
    )
    color_map = Array(dtype=np.uint32).tag(
        sync=True, **array_serialization_wrap("color_map")
    )
    wireframe = Bool().tag(sync=True)
    outlines = Bool().tag(sync=True)
    outlines_color = Int(min=0, max=0xFFFFFF).tag(sync=True)
    opacity = TimeSeries(Float(min=0.0, max=1.0, default_value=1.0)).tag(sync=True)
    model_matrix = TimeSeries(Array(dtype=np.float32)).tag(
        sync=True, **array_serialization_wrap("model_matrix")
    )

    def __init__(self, **kwargs):
        super(VoxelsGroup, self).__init__(**kwargs)

        self.set_trait("type", "VoxelsGroup")

    def get_bounding_box(self):
        return get_bounding_box(self.model_matrix)



import k3d
import pytest
from .plot_compare import *


def test_text():
    prepare()

    pytest.plot += k3d.text(
        '\\int_{-\\infty}^\\infty \\hat f(\\xi)\\,e^{2 \\pi i \\xi x} \\,d\\xi',
        [0.75, 0.5, 0.0], color=0, size=1.5, reference_point='lt')

    pytest.plot += k3d.text('{(1,1,\\frac{5}{\\pi})}', [0.25, 0.5, 1.0],
                            color=0, size=1.5, reference_point='rb')
    pytest.plot += k3d.text('<h1 style="color: red;">Important!</h1>Hello<strong>World</strong>',
                            [-0.5, 0.5, 0.5],
                            color=0, size=1.5, is_html=True, reference_point='rb')

    compare('text', False)


def test_text_without_box():
    prepare()

    pytest.plot += k3d.text(
        '\\int_{-\\infty}^\\infty \\hat f(\\xi)\\,e^{2 \\pi i \\xi x} \\,d\\xi',
        [0.75, 0.5, 0.0], color=0, size=1.5, reference_point='lt', label_box=False)

    pytest.plot += k3d.text('{(1,1,\\frac{5}{\\pi})}', [0.25, 0.5, 1.0],
                            color=0, size=1.5, reference_point='rb', label_box=False)
    pytest.plot += k3d.text('<h1 style="color: red;">Important!</h1>Hello<strong>World</strong>',
                            [-0.5, 0.5, 0.5],
                            color=0, size=1.5, is_html=True, reference_point='rb',
                            label_box=False)

    compare('text_without_box', False)



import k3d
import numpy as np
import pytest
from .plot_compare import *

T = 1.618033988749895
r = 4.77
zmin, zmax = -r, r
xmin, xmax = -r, r
ymin, ymax = -r, r
Nx, Ny, Nz = 77, 77, 77

x = np.linspace(xmin, xmax, Nx)
y = np.linspace(ymin, ymax, Ny)
z = np.linspace(zmin, zmax, Nz)
x, y, z = np.meshgrid(x, y, z, indexing='ij')
p = 2 - (np.cos(x + T * y) + np.cos(x - T * y) + np.cos(y + T * z) + np.cos(y - T * z) + np.cos(
    z - T * x) + np.cos(z + T * x)).astype(np.float32)


def test_marching_cubes():
    global p
    prepare()

    iso = k3d.marching_cubes(p, level=0.0)

    pytest.plot += iso

    compare('marching_cubes')

    iso.level = 0.1

    compare('marching_cubes_dynamic_level')


def test_marching_cubes_smoothed():
    global p
    prepare()

    iso = k3d.marching_cubes(p, level=0.0, flat_shading=False)

    pytest.plot += iso

    compare('marching_cubes_smoothed')


def test_marching_cubes_wireframe():
    global p
    prepare()

    iso = k3d.marching_cubes(p, level=0.0, wireframe=True)

    pytest.plot += iso

    compare('marching_cubes_wireframe')


def test_marching_cubes_non_uniformly_spaced():
    prepare()

    ax = np.logspace(0, np.log10(xmax - xmin + 1), Nx, endpoint=True) + xmin - 1
    ay = np.logspace(0, np.log10(ymax - ymin + 1), Nx, endpoint=True) + ymin - 1
    az = np.logspace(0, np.log10(zmax - zmin + 1), Nx, endpoint=True) + zmin - 1

    x, y, z = np.meshgrid(ax, ay, az, indexing='ij')
    p = 2 - (np.cos(x + T * y) + np.cos(x - T * y) + np.cos(y + T * z) + np.cos(y - T * z) + np.cos(
        z - T * x) + np.cos(z + T * x)).astype(np.float32)

    iso = k3d.marching_cubes(p, level=0.0, color=1,
                             spacings_x=(ax[1:] - ax[:-1]).astype(np.float32),
                             spacings_y=(ay[1:] - ay[:-1]).astype(np.float32),
                             spacings_z=(az[1:] - az[:-1]).astype(np.float32),
                             bounds=[xmin, xmax, ymin, ymax, zmin, zmax],
                             opacity=0.15,
                             wireframe=True)
    pytest.plot += iso

    compare('marching_cubes_non_uniformly_spaced')



import k3d
import pytest
from .plot_compare import *


def test_texture_text():
    prepare()

    position1 = [0.5, 1.2, 0.0]
    position2 = [-0.5, -1.2, 0.0]

    pytest.plot += k3d.texture_text('K3D Jupyter', position1, color=0xff0000, font_face='Arial',
                                    font_weight=300, size=2.0)
    pytest.plot += k3d.texture_text('K3D Jupyter', position2, color=0xff00ff, font_face='Arial',
                                    font_weight=700, size=1.0)

    compare('texture_text', camera_factor=0.5)



import k3d
import numpy as np
import pytest
from .plot_compare import *
import math

v = []
s = []
o = []

for i in range(10):
    for fi in np.arange(0, 2 * math.pi, 0.1):
        v.append([
            math.sin(fi * 4 + i),
            math.cos(fi * 7 + i),
            math.cos(fi * 3) + fi * 1.5
        ])
        s.append(math.sin(fi * i))
        o.append(abs(math.cos(fi * i)))

v = np.array(v, dtype=np.float32)
s = np.array(s, dtype=np.float32)
o = np.array(o, dtype=np.float32)


def test_points_flat():
    global v, s, o

    prepare()

    points = k3d.points(v, shader='flat', point_size=0.1, opacities=o,
                        color=0xff0000)

    pytest.plot += points

    compare('points_flat')


def test_points_3d():
    global v, s, o

    prepare()

    points = k3d.points(v, shader='3d', point_size=0.2, opacities=o,
                        color=0xff0000)

    pytest.plot += points

    compare('points_3d')


def test_points_3d_clipping_plane():
    global v, s, o

    prepare()

    points = k3d.points(v, shader='3d', point_size=0.2, opacities=o,
                        color=0xff0000)
    pytest.plot.clipping_planes = [
        [1, 1, 0, 0]
    ]
    pytest.plot += points

    compare('points_3d_clipping_plane')


def test_points_3dSpecular():
    global v, s, o

    prepare()

    points = k3d.points(v, shader='3dSpecular', point_size=0.2, opacities=o,
                        color=0xff0000)

    pytest.plot += points

    compare('points_3dSpecular')


def test_points_3dSpecular_sizes():
    global v, s, o

    prepare()

    points = k3d.points(v, shader='3dSpecular', opacities=o,
                        point_sizes=np.linspace(0, 0.2, v.shape[0]), color=0xff0000)

    pytest.plot += points

    compare('points_3dSpecular_sizes')


def test_points_mesh_sizes():
    global v, s, o

    prepare()

    points = k3d.points(v, shader='mesh', opacities=o,
                        point_sizes=np.linspace(0, 0.2, v.shape[0]), color=0xff0000)

    pytest.plot += points

    compare('points_mesh_sizes')


def test_points_mesh():
    global v, s, o

    prepare()

    points = k3d.points(v, shader='mesh', point_size=0.2, opacities=o,
                        color=0xff0000)

    pytest.plot += points

    compare('points_mesh')


def test_points_mesh_clipping_plane():
    global v, s, o

    prepare()

    points = k3d.points(v, shader='mesh', point_size=0.2, opacities=o,
                        color=0xff0000)
    pytest.plot.clipping_planes = [
        [1, 1, 0, 0]
    ]
    pytest.plot += points

    compare('points_mesh_clipping_plane')


def test_points_mesh_low_detail():
    prepare()

    points = k3d.points(np.array([[0, 0, 0], [1, 0, 0]]), shader='mesh', point_size=0.3,
                        mesh_detail=1, color=0xff0000)

    pytest.plot += points

    compare('points_mesh_low_detail')


def test_points_mesh_high_detail():
    prepare()

    points = k3d.points(np.array([[0, 0, 0], [1, 0, 0]]), shader='mesh', point_size=0.3,
                        mesh_detail=8, color=0xff0000)

    pytest.plot += points

    compare('points_mesh_high_detail')


def test_points_dot():
    global v, s

    prepare()

    points = k3d.points(v, shader='dot', point_size=3, opacities=o,
                        color=0xff0000)

    pytest.plot += points

    compare('points_dot')



import k3d
import pytest
from .plot_compare import *
import numpy as np

width = height = 50


def f(x, y):
    return np.sin(float(x) / width * np.pi * 2.0), np.cos(float(y) / height * np.pi * 2.0)


colors = np.array([[(0xFF0000, 0x00FF00) for x in range(width)] for y in range(height)])
vectors = np.array([[f(x, y) for x in range(width)] for y in range(height)])


def test_vector_field():
    global vectors, colors, model_matrix

    prepare()

    vector_field = k3d.vector_field(vectors[::5, ::5], colors[::5, ::5], scale=1.5,
                                    line_width=0.001, head_size=1.5,
                                    transform=k3d.transform(rotation=[np.pi / 2, 1, 0, 0]))

    pytest.plot += vector_field

    compare('vector_field')


def test_vector_field_no_head():
    global vectors, colors, model_matrix

    prepare()

    vector_field = k3d.vector_field(vectors, colors, scale=1.5, use_head=False, line_width=0.001,
                                    transform=k3d.transform(rotation=[np.pi / 2, 1, 0, 0]))

    pytest.plot += vector_field

    compare('vector_field_no_head')


def test_vector_field_scale():
    global vectors, colors, model_matrix

    prepare()

    vector_field = k3d.vector_field(vectors, colors, scale=2.5, line_width=0.001,
                                    transform=k3d.transform(rotation=[np.pi / 2, 1, 0, 0]))

    pytest.plot += vector_field

    compare('vector_field_scale')



from ._version import version_info, __version__

from .colormaps import paraview_color_maps
from .colormaps import basic_color_maps
from .colormaps import matplotlib_color_maps

from .factory import (plot,
                      nice_colors,
                      line,
                      marching_cubes,
                      mesh,
                      points,
                      stl,
                      surface,
                      text,
                      text2d,
                      texture,
                      texture_text,
                      vector_field,
                      vectors,
                      voxels,
                      voxels_group,
                      sparse_voxels,
                      volume,
                      mip,
                      label,
                      vtk_poly_data,
                      voxel_chunk)

from .plot import Plot

from .transfer_function_editor import transfer_function_editor

from .transform import transform

def _jupyter_nbextension_paths():
    return [{
        'section': 'notebook',
        'src': 'static',
        'dest': 'k3d',
        'require': 'k3d/extension'
    }]



import k3d
import numpy as np
import pytest
from .plot_compare import *
import vtk
from vtk.util import numpy_support


def test_mip():
    prepare()

    reader = vtk.vtkXMLImageDataReader()
    reader.SetFileName('./test/assets/volume.vti')
    reader.Update()
    vti = reader.GetOutput()

    x, y, z = vti.GetDimensions()
    volume_data = numpy_support.vtk_to_numpy(
        vti.GetPointData().GetArray(0)
    ).reshape(-1, y, x).astype(np.float32)

    volume = k3d.mip(volume_data, samples=512)

    pytest.plot += volume

    compare('mip')


def test_mip_opacity_function():
    prepare()

    reader = vtk.vtkXMLImageDataReader()
    reader.SetFileName('./test/assets/volume.vti')
    reader.Update()
    vti = reader.GetOutput()

    x, y, z = vti.GetDimensions()
    volume_data = numpy_support.vtk_to_numpy(
        vti.GetPointData().GetArray(0)
    ).reshape(-1, y, x).astype(np.float32)

    volume = k3d.mip(volume_data, opacity_function=[0, 0.0, 0.2, 0.5, 1, 1.0], samples=512)

    pytest.plot += volume

    compare('mip_opacity_function')



import os, sys, inspect

current_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parent_dir = os.path.dirname(current_dir)
sys.path.insert(0, parent_dir)

import pytest
import k3d
from k3d.headless import k3d_remote, get_headless_driver
import subprocess
import os


def pytest_configure(config):
    """
    Allows plugins and conftest files to perform initial configuration.
    This hook is called for every plugin and initial conftest
    file after command line options have been parsed.
    """
    process = subprocess.Popen('webpack', cwd=os.path.abspath('./../js/'), shell=True)
    process.wait()


def pytest_sessionstart(session):
    """
    Called after the Session object has been created and
    before performing collection and entering the run test loop.
    """
    pytest.plot = k3d.plot(screenshot_scale=1.0, antialias=2, camera_auto_fit=False,
                           colorbar_object_id=0)
    pytest.headless = k3d_remote(pytest.plot, get_headless_driver(), port=8090)
    pytest.headless.browser.execute_script("window.randomMul = 0.0;")


def pytest_sessionfinish(session, exitstatus):
    """
    Called after whole test run finished, right before
    returning the exit status to the system.
    """

    pytest.headless.close()


def pytest_unconfigure(config):
    """
    called before test process is exited.
    """



import k3d
import pytest
from .plot_compare import *
import numpy as np
from math import sqrt, sin, cos

color_map = (0xffff00, 0xff0000, 0x00ff00)


def r(x, y, z, width=100, height=100, length=100):
    r = sqrt((x - width / 2) * (x - width / 2) + (y - height / 2) * (y - height / 2) + (
            z - length / 2) * (z - length / 2))
    r += sin(x / 2) * 3
    r += cos(y / 10) * 5

    return r


def f(x, y, z, width=100, height=100, length=100):
    return 0 if r(x, y, z, width, height, length) > width / 2 else (
        1 if z + sin(x / 20) * 10 > length / 2 else 2)


voxels = np.array(
    [[[f(x, y, z, 100, 100, 100) for x in range(100)] for y in range(100)] for z in range(100)])


def test_voxels():
    global color_map, voxels

    prepare()

    obj = k3d.voxels(voxels.astype(np.uint8), color_map, outlines=False)

    pytest.plot += obj

    compare('voxels')

def test_voxels_sparse():
    global color_map, voxels

    prepare()

    sparse_data = []

    for val in np.unique(voxels):
        if val != 0:
            z, y, x = np.where(voxels==val)
            sparse_data.append(np.dstack((x, y, z, np.full(x.shape, val))).reshape(-1,4).astype(np.uint16))

    sparse_data = np.vstack(sparse_data)
    obj = k3d.sparse_voxels(sparse_data, (100, 100, 100), color_map, outlines=False)

    pytest.plot += obj

    compare('voxels_sparse')

def test_voxels_wireframe():
    global color_map, voxels

    prepare()

    obj = k3d.voxels(voxels.astype(np.uint8), color_map, outlines=False, wireframe=True)

    pytest.plot += obj

    compare('voxels_wireframe')


def test_voxels_outline():
    global color_map, voxels

    prepare()

    obj = k3d.voxels(voxels.astype(np.uint8), color_map, outlines=True)

    pytest.plot += obj

    compare('voxels_outline')

    obj.outlines_color = 0xff0000

    compare('voxels_outline_dynamic_outlines_color')


def test_voxels_outline_opacity():
    global color_map, voxels

    prepare()

    obj = k3d.voxels(voxels.astype(np.uint8), color_map, outlines=True, opacity=0.5)

    pytest.plot += obj

    compare('voxels_outline_opacity')


def test_voxels_outline_clipping_plane():
    global color_map, voxels

    prepare()

    obj = k3d.voxels(voxels.astype(np.uint8), color_map, outlines=True)

    pytest.plot += obj
    pytest.plot.clipping_planes = [
        [-1, 0, -1, 50]
    ]

    compare('voxels_outline_clipping_plane')


def test_voxels_box():
    global color_map, voxels

    prepare()

    width, height, length = 200, 50, 100
    box_voxels = voxels = np.array(
        [[[f(x, y, z, width, height, length) for x in range(width)] for y in range(height)] for z in
         range(length)])

    obj = k3d.voxels(box_voxels.astype(np.uint8), color_map, outlines=True)

    pytest.plot += obj

    compare('voxels_box')




import numpy as np
from itertools import product
import k3d


class PlatonicSolid(object):
    """Base class for platonic solids."""
    
    @property
    def mesh(self):
        """Return k3d.mesh object of the solid."""
        return k3d.mesh(self.vertices, self.indices, side='double')

    @property
    def points(self):
        """Return k3d.points object of solid vertices."""
        return k3d.points(self.vertices, point_size=np.sqrt(np.sum((self.vertices[0] - self.vertices[1]) ** 2)) / 20)

class Dodecahedron(PlatonicSolid):
    """Create a dodecahedron solid.
        
    Arguments:
        origin: `array_like`. 
            The position of centroid of the solid.
        size: `float`.
            The size*sqrt(3) is the distance of each vertex of the solid.
    """
    def __init__(self, origin=[0, 0, 0], size=1):
        origin = np.array(origin, dtype=np.float32)

        if origin.shape == (3,):
            fi = (1 + np.sqrt(5)) / 2
            dodecahedron_vertices = list(product([-1, 1], [-1, 1], [-1, 1]))
            dodecahedron_vertices += [(0, fi, 1 / fi), (0, -fi, 1 / fi), (0, -fi, -1 / fi), (0, fi, -1 / fi),
                                      (1 / fi, 0, fi), (-1 / fi, 0, fi), (-1 / fi, 0, -fi), (1 / fi, 0, -fi),
                                      (fi, 1 / fi, 0), (-fi, 1 / fi, 0), (-fi, -1 / fi, 0), (fi, -1 / fi, 0)]
            dodecahedron_vertices = np.float32(size * np.array(dodecahedron_vertices) + origin)

            self.vertices = dodecahedron_vertices
            self.indices = [0, 1, 18, 0, 1, 10, 1, 9, 10, 0, 10, 14, 10, 14, 15, 4, 10, 15, 4, 9, 10, 4, 5, 9, 4, 5, 19,
                            4, 15, 19, 6, 15, 19, 6, 16, 19,
                            6, 7, 16, 6, 7, 8, 6, 8, 11, 2, 3, 17, 2, 3, 8, 2, 8, 11, 1, 3, 13, 1, 3, 18, 3, 17, 18, 1,
                            9, 13, 9, 12, 13, 5, 9, 12,
                            5, 12, 19, 12, 16, 19, 7, 12, 16, 3, 7, 8, 3, 7, 12, 3, 12, 13, 14, 6, 15, 14, 6, 11, 2, 11,
                            14, 0, 17, 18, 0, 2, 17, 0, 2, 14]
        else:
            raise TypeError('Origin should have 3 coordinates.')


class Cube(PlatonicSolid):
    """Create a cube.
        
    Arguments:
        origin: `array_like`. 
            The position of centroid of the solid.
        size: `float`.
            The size*sqrt(3) is the distance of each vertex from centroid of the solid.
            
            The edge of cube is 2*size.
    """
    def __init__(self, origin=[0, 0, 0], size=1):
        origin = np.array(origin, dtype=np.float32)

        if origin.shape == (3,):
            cube_vertices = np.array(list(product([1, -1], [1, -1], [1, -1])), np.float32)
            cube_vertices = np.float32(size * cube_vertices + origin)

            self.vertices = cube_vertices
            self.indices = [0, 1, 2, 1, 2, 3, 0, 1, 4, 1, 4, 5, 1, 3, 5, 3, 5, 7, 0, 2, 4, 2, 4, 6, 2, 3, 7, 2, 6, 7, 4,
                            5, 6, 5, 6, 7]

        else:
            raise TypeError('Origin attribute should have 3 coordinates.')


class Icosahedron(PlatonicSolid):
    """Create a icosahedron solid.
        
    Arguments:
        origin: `array_like`. 
            The position of centroid of the solid.
        size: `float`.
            The size of the solid.
    """
    def __init__(self, origin=[0, 0, 0], size=1):
        origin = np.array(origin, dtype=np.float32)

        if origin.shape == (3,):
            fi = (1 + np.sqrt(5)) / 2
            icosahedron_vertices = [(1, fi, 0), (1, -fi, 0), (-1, fi, 0), (-1, -fi, 0),
                                    (fi, 0, 1), (fi, 0, -1), (-fi, 0, 1), (-fi, 0, -1),
                                    (0, 1, fi), (0, 1, -fi), (0, -1, fi), (0, -1, -fi)]
            icosahedron_vertices = np.float32(size * np.array(icosahedron_vertices) + origin)

            self.vertices = icosahedron_vertices
            self.indices = [0, 2, 8, 0, 4, 8, 0, 2, 9, 0, 5, 9, 2, 6, 8, 2, 7, 9, 2, 6, 7, 0, 4, 5, 1, 4, 5, 1, 5, 11,
                            7, 9, 11, 3, 7, 11, 3, 6, 7, 3, 6, 10, 4, 8, 10, 6, 8, 10, 1, 4, 10, 1, 3, 11, 1, 3, 10, 5,
                            9, 11]

        else:
            raise TypeError('Origin attribute should have 3 coordinates.')


class Octahedron(PlatonicSolid):
    """Create a octahedron solid.
        
    Arguments:
        origin: `array_like`. 
            The position of centroid of the solid.
        size: `float`.
            The distance of each vertex from centroid of the solid.
    """
    def __init__(self, origin=[0, 0, 0], size=1):
        origin = np.array(origin, dtype=np.float32)

        if origin.shape == (3,):
            octahedron_vertices = [(1, 0, 0), (0, 1, 0), (0, 0, 1), (-1, 0, 0), (0, -1, 0), (0, 0, -1)]
            octahedron_vertices = np.float32(size * np.array(octahedron_vertices) + origin)

            self.vertices = octahedron_vertices
            self.indices = [0, 1, 2, 0, 1, 5, 1, 2, 3, 1, 3, 5, 0, 4, 5, 0, 2, 4, 2, 3, 4, 3, 4, 5]

        else:
            raise TypeError('Origin attribute should have 3 coordinates.')


class Tetrahedron(PlatonicSolid):
    """Create a tetrahedron solid.
        
    Arguments:
        origin: `array_like`. 
            The position of centroid of the solid.
        size: `float`.
            The size*sqrt(3) is distance of each vertex of the solid.
    """
    def __init__(self, origin=[0, 0, 0], size=1):
        origin = np.array(origin, dtype=np.float32)

        if origin.shape == (3,):
            tetrahedron_vertices = [(1, 1, 1), (1, -1, -1), (-1, 1, -1), (-1, -1, 1)]
            tetrahedron_vertices = np.float32(size * np.array(tetrahedron_vertices) + origin)

            self.vertices = tetrahedron_vertices
            self.indices = [0, 1, 2, 0, 1, 3, 1, 2, 3, 0, 2, 3]

        else:
            raise TypeError('Origin attribute should have 3 coordinates.')




import k3d
import numpy as np
import pytest
from .plot_compare import *
import math
from k3d.helpers import download
import vtk
from vtk.util import numpy_support


def test_texture():
    prepare()

    # Y-Z flip
    model_matrix = [
        1.0, 0.0, 0.0, 0.0,
        0.0, -1.0, 0.0, 0.0,
        0.0, 0.0, 1.0, 0.0,
        0.0, 0.0, 0.0, 1.0
    ]

    texture = k3d.texture(open('./test/assets/texture.png', 'br').read(), 'png',
                          rotation=[math.radians(90), 1, 0, 0],
                          model_matrix=model_matrix,
                          name='Photo')

    pytest.plot += texture

    compare('texture')

    texture.binary = open('./test/assets/mandelbrot.jpg', 'br').read()
    texture.name = 'Fractal'

    compare('texture_dynamic_change')


def test_texture_attribute():
    prepare()

    reader = vtk.vtkXMLImageDataReader()
    reader.SetFileName('./test/assets/volume.vti')
    reader.Update()
    vti = reader.GetOutput()

    x, y, z = vti.GetDimensions()
    volume_data = numpy_support.vtk_to_numpy(
        vti.GetPointData().GetArray(0)
    ).reshape(-1, y, x).astype(np.float32)

    texture = k3d.texture(attribute=volume_data[64],
                          transform=k3d.transform(rotation=[np.pi / 2, 1, 0, 0]),
                          color_map=k3d.basic_color_maps.Jet,
                          opacity_function=[-1, 1, 1, 1],
                          name='Slice')

    pytest.plot += texture

    compare('texture_attribute')

    texture.interpolation = False

    compare('texture_attribute_dynamic_interpolation')



from __future__ import print_function

import os

from setuptools import setup, find_packages

from distutils import log

from setupbase import (
    create_cmdclass, install_npm, combine_commands,
    ensure_targets, skip_npm
)

here = os.path.dirname(os.path.abspath(__file__))
node_root = os.path.join(here, 'js')

log.set_verbosity(log.DEBUG)
log.info('setup.py entered')
log.info('$PATH=%s' % os.environ['PATH'])

LONG_DESCRIPTION = 'Jupyter notebook extension for 3D visualization.'

# Representative files that should exist after a successful build
targets = [
    os.path.join(here, 'k3d', 'static', 'standalone.js'),
    os.path.join(here, 'k3d', 'static', 'snapshot_standalone.txt'),
    os.path.join(here, 'k3d', 'static', 'snapshot_online.txt'),
    os.path.join(here, 'k3d', 'static', 'snapshot_inline.txt'),
    os.path.join(here, 'k3d', 'static', 'headless.html'),
    os.path.join(here, 'k3d', 'static', 'extension.js'),
    os.path.join(here, 'k3d', 'static', 'index.js')
]

cmdclass = create_cmdclass(
    wrappers=[] if skip_npm else ['jsdeps'],
    data_dirs=[]
)
cmdclass['jsdeps'] = combine_commands(
    install_npm(node_root),
    ensure_targets(targets)
)

version_ns = {}
with open(os.path.join(here, 'k3d', '_version.py')) as f:
    exec (f.read(), {}, version_ns)

setup_args = {
    'name': 'K3D',
    'version': version_ns['__version__'],
    'license': 'MIT',
    'description': 'Jupyter notebook extension for 3D visualization.',
    'long_description': LONG_DESCRIPTION,
    'include_package_data': True,
    'data_files': [
        (
            'share/jupyter/nbextensions/k3d',
            [
                'k3d/static/extension.js',
                'k3d/static/snapshot_standalone.txt',
                'k3d/static/snapshot_online.txt',
                'k3d/static/snapshot_inline.txt',
                'k3d/static/headless.html',
                'k3d/static/standalone.js',
                'k3d/static/index.js',
                'k3d/static/index.js.map',
            ]
        ),
        (
            'etc/jupyter/nbconfig/notebook.d',
            [
                'k3d/k3d_extension.json',
            ]
        ),
    ],
    'install_requires': [
        'ipywidgets>=7.0.1',
        'ipydatawidgets',
        'traittypes',
        'traitlets',
        'numpy>=1.11.0',
        'msgpack'
    ],
    'packages': find_packages(),
    'zip_safe': False,
    'cmdclass': cmdclass,
    'author': 'k3d team',
    'author_email': 'artur.trzesiok@gmail.com',
    'url': 'http://jupyter.org',
    'keywords': [
        'ipython',
        'jupyter',
        'widgets',
    ],
    'classifiers': [
        'Development Status :: 4 - Beta',
        'Framework :: IPython',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Multimedia :: Graphics',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
}

if __name__ == '__main__':
    setup(**setup_args)



import k3d
import pytest
from .plot_compare import *


def test_text2d():
    prepare()

    pytest.plot += k3d.text2d(
        '\\int_{-\\infty}^\\infty \\hat f(\\xi)\\,e^{2 \\pi i \\xi x} \\,d\\xi',
        [0.75, 0.5], color=0, size=1.5, reference_point='lt')

    pytest.plot += k3d.text2d('{(1,1,\\frac{5}{\\pi})}', [0.25, 0.5],
                              color=0, size=1.5, reference_point='rb')
    pytest.plot += k3d.text2d('<h1 style="color: red;">Important!</h1>Hello<strong>World</strong>',
                              [0.5, 0.5],
                              color=0, size=1.5, is_html=True, reference_point='rb')

    compare('text2d', False)


def test_text2d_without_box():
    prepare()

    pytest.plot += k3d.text2d(
        '\\int_{-\\infty}^\\infty \\hat f(\\xi)\\,e^{2 \\pi i \\xi x} \\,d\\xi',
        [0.75, 0.5], color=0, size=1.5, reference_point='lt', label_box=False)

    pytest.plot += k3d.text2d('{(1,1,\\frac{5}{\\pi})}', [0.25, 0.5],
                              color=0, size=1.5, reference_point='rb', label_box=False)
    pytest.plot += k3d.text2d('<h1 style="color: red;">Important!</h1>Hello<strong>World</strong>',
                              [0.5, 0.5],
                              color=0, size=1.5, is_html=True, reference_point='rb',
                              label_box=False)

    compare('text2d_without_box', False)



from __future__ import print_function

import base64
import ipywidgets as widgets
from IPython.display import display
from functools import wraps
from traitlets import Unicode, Bool, Int, List, Float

from ._version import __version__ as version
from .objects import (Line, Label, MIP, MarchingCubes, Mesh, Points, STL, SparseVoxels, Surface,
                      Text, Text2d, Texture, TextureText, VectorField, Vectors, Volume, Voxels,
                      VoxelsGroup, ListOrArray, Drawable, TimeSeries)

objects_map = {
    'Line': Line,
    'Label': Label,
    'MIP': MIP,
    'Marchingcubes': MarchingCubes,
    'Mesh': Mesh,
    'Points': Points,
    'STL': STL,
    'SparseVoxels': SparseVoxels,
    'Surface': Surface,
    'Text': Text,
    'Text2d': Text2d,
    'Texture': Texture,
    'TextureText': TextureText,
    'VectorField': VectorField,
    'Vectors': Vectors,
    'Volume': Volume,
    'Voxels': Voxels,
    'VoxelsGroup': VoxelsGroup
}
import numpy as np


class Plot(widgets.DOMWidget):
    """
    Main K3D widget.

    Attributes:
        antialias: `int`:
            Enable antialiasing in WebGL renderer, changes have no effect after displaying.
        height: `int`:
            Height of the Widget in pixels, changes have no effect after displaying.
        background_color: `int`.
            Packed RGB color of the plot background (0xff0000 is red, 0xff is blue), -1 is for transparent.
        camera_auto_fit: `bool`.
            Enable automatic camera setting after adding, removing or changing a plot object.
        grid_auto_fit: `bool`.
            Enable automatic adjustment of the plot grid to contained objects.
        grid_color: `int`.
            Packed RGB color of the plot grids (0xff0000 is red, 0xff is blue).
        grid_visible: `bool`.
            Enable or disable grid.
        screenshot_scale: `Float`.
            Multipiler to screenshot resolution.
        voxel_paint_color: `int`.
            The (initial) int value to be inserted when editing voxels.
        label_color: `int`.
            Packed RGB color of the labels (0xff0000 is red, 0xff is blue).
        lighting: `Float`.
            Lighting factor.
        grid: `array_like`.
            6-element tuple specifying the bounds of the plot grid (x0, y0, z0, x1, y1, z1).
        camera: `array_like`.
            9-element list or array specifying camera position.
        camera_no_rotate: `Bool`.
            Lock for camera rotation.
        camera_no_zoom: `Bool`.
            Lock for camera zoom.
        camera_no_pan: `Bool`.
            Lock for camera pan.
        camera_rotate_speed: `Float`.
            Speed of camera rotation.
        camera_zoom_speed: `Float`.
            Speed of camera zoom.
        camera_pan_speed: `Float`.
            Speed of camera pan.
        camera_fov: `Float`.
            Camera Field of View.
        camera_damping_factor: `Float`.
            Defines the intensity of damping. Default is 0 (disabled).
        snapshot_type: `string`.
            Can be 'full', 'online' or 'inline'.
        axes: `list`.
            Axes labels for plot.
        time: `list`.
            Time value (used in TimeSeries)
        name: `string`.
            Name of the plot. Used to filenames of snapshot/screenshot etc.
        mode: `str`.
            Mode of K3D viewer.

            Legal values are:

            :`view`: No interaction with objects,

            :`add`: On voxels objects adding mode,

            :`change`: On voxels objects edit mode,

            :`callback`: Handling click_callback and hover_callback on some type of objects,

            :`manipulate`: Enable object transform widget.
        camera_mode: `str`.
            Mode of camera movement.

            Legal values are:

            :`trackball`: orbit around point with dynamic up-vector of camera,

            :`orbit`: orbit around point with fixed up-vector of camera,

            :`fly`: orbit around point with dynamic up-vector of camera, mouse wheel also moves target point.
        manipulate_mode: `str`.
            Mode of manipulate widgets.

            Legal values are:

            :`translate`: Translation widget,

            :`rotate`: Rotation widget,

            :`scale`: Scaling widget.
        auto_rendering: `Bool`.
            State of auto rendering.
        fps: `Float`.
            Fps of animation.
        objects: `list`.
            List of `k3d.objects.Drawable` currently included in the plot, not to be changed directly.
    """

    _view_name = Unicode("PlotView").tag(sync=True)
    _model_name = Unicode("PlotModel").tag(sync=True)
    _view_module = Unicode("k3d").tag(sync=True)
    _model_module = Unicode("k3d").tag(sync=True)

    _view_module_version = Unicode(version).tag(sync=True)
    _model_module_version = Unicode(version).tag(sync=True)
    _backend_version = Unicode(version).tag(sync=True)

    # readonly (specified at creation)
    antialias = Int(min=0, max=5).tag(sync=True)
    height = Int().tag(sync=True)

    # readonly (not to be modified directly)
    object_ids = List().tag(sync=True)

    # read-write
    camera_auto_fit = Bool(True).tag(sync=True)
    auto_rendering = Bool(True).tag(sync=True)
    lighting = Float().tag(sync=True)
    fps = Float().tag(sync=True)
    grid_auto_fit = Bool(True).tag(sync=True)
    grid_visible = Bool(True).tag(sync=True)
    fps_meter = Bool(True).tag(sync=True)
    menu_visibility = Bool(True).tag(sync=True)
    screenshot_scale = Float().tag(sync=True)
    time = Float().tag(sync=True)
    grid = ListOrArray((-1, -1, -1, 1, 1, 1), minlen=6, maxlen=6).tag(sync=True)
    grid_color = Int().tag(sync=True)
    label_color = Int().tag(sync=True)
    background_color = Int().tag(sync=True)
    voxel_paint_color = Int().tag(sync=True)
    camera = ListOrArray(minlen=9, maxlen=9, empty_ok=True).tag(sync=True)
    camera_animation = TimeSeries(List()).tag(sync=True)
    camera_no_rotate = Bool(False).tag(sync=True)
    camera_no_zoom = Bool(False).tag(sync=True)
    camera_no_pan = Bool(False).tag(sync=True)
    camera_rotate_speed = Float().tag(sync=True)
    camera_zoom_speed = Float().tag(sync=True)
    camera_pan_speed = Float().tag(sync=True)
    camera_damping_factor = Float().tag(sync=True)
    clipping_planes = ListOrArray(empty_ok=True).tag(sync=True)
    colorbar_object_id = Int(-1).tag(sync=True)
    colorbar_scientific = Bool(False).tag(sync=True)
    rendering_steps = Int(1).tag(sync=True)
    screenshot = Unicode().tag(sync=True)
    snapshot = Unicode().tag(sync=True)
    snapshot_type = Unicode().tag(sync=True)
    camera_fov = Float().tag(sync=True)
    name = Unicode(default_value=None, allow_none=True).tag(sync=True)
    axes = List(minlen=3, maxlen=3, default_value=["x", "y", "z"]).tag(sync=True)
    axes_helper = Float().tag(sync=True)
    mode = Unicode().tag(sync=True)
    camera_mode = Unicode().tag(sync=True)
    manipulate_mode = Unicode().tag(sync=True)

    objects = []

    def __init__(
            self,
            antialias=3,
            background_color=0xFFFFFF,
            camera_auto_fit=True,
            grid_auto_fit=True,
            grid_visible=True,
            height=512,
            voxel_paint_color=0,
            grid=(-1, -1, -1, 1, 1, 1),
            screenshot_scale=2.0,
            lighting=1.5,
            time=0.0,
            fps_meter=False,
            menu_visibility=True,
            colorbar_object_id=-1,
            rendering_steps=1,
            axes=["x", "y", "z"],
            camera_no_rotate=False,
            camera_no_zoom=False,
            camera_rotate_speed=1.0,
            camera_zoom_speed=1.2,
            camera_pan_speed=0.3,
            snapshot_type='full',
            camera_no_pan=False,
            camera_fov=45.0,
            camera_damping_factor=0.0,
            axes_helper=1.0,
            name=None,
            mode="view",
            camera_mode="trackball",
            manipulate_mode="translate",
            auto_rendering=True,
            fps=25.0,
            grid_color=0xE6E6E6,
            label_color=0x444444,
            *args,
            **kwargs
    ):
        super(Plot, self).__init__()

        self.antialias = antialias
        self.camera_auto_fit = camera_auto_fit
        self.grid_auto_fit = grid_auto_fit
        self.fps_meter = fps_meter
        self.fps = fps
        self.grid = grid
        self.grid_visible = grid_visible
        self.background_color = background_color
        self.grid_color = grid_color
        self.label_color = label_color
        self.voxel_paint_color = voxel_paint_color
        self.screenshot_scale = screenshot_scale
        self.height = height
        self.lighting = lighting
        self.time = time
        self.menu_visibility = menu_visibility
        self.colorbar_object_id = colorbar_object_id
        self.rendering_steps = rendering_steps
        self.camera_no_rotate = camera_no_rotate
        self.camera_no_zoom = camera_no_zoom
        self.camera_no_pan = camera_no_pan
        self.camera_rotate_speed = camera_rotate_speed
        self.camera_zoom_speed = camera_zoom_speed
        self.camera_pan_speed = camera_pan_speed
        self.camera_damping_factor = camera_damping_factor
        self.camera_fov = camera_fov
        self.axes = axes
        self.axes_helper = axes_helper
        self.name = name
        self.mode = mode
        self.snapshot_type = snapshot_type
        self.camera_mode = camera_mode
        self.manipulate_mode = manipulate_mode
        self.auto_rendering = auto_rendering
        self.camera = []

        self.object_ids = []
        self.objects = []

        self.outputs = []

    def __iadd__(self, objs):
        """Add Drawable to plot."""
        assert isinstance(objs, Drawable)

        for obj in objs:
            if obj.id not in self.object_ids:
                self.object_ids = self.object_ids + [obj.id]
                self.objects.append(obj)

        return self

    def __isub__(self, objs):
        """Remove Drawable from plot."""
        assert isinstance(objs, Drawable)

        for obj in objs:
            self.object_ids = [id_ for id_ in self.object_ids if id_ != obj.id]
            if obj in self.objects:
                self.objects.remove(obj)

        return self

    def display(self, **kwargs):
        """Show plot inside ipywidgets.Output()."""
        output = widgets.Output()

        with output:
            display(self, **kwargs)

        self.outputs.append(output)

        display(output)

    def render(self):
        """Trigger rendering on demand.

        Useful when self.auto_rendering == False."""
        self.send({"msg_type": "render"})

    def start_auto_play(self):
        """Start animation of plot with objects using TimeSeries."""
        self.send({"msg_type": "start_auto_play"})

    def stop_auto_play(self):
        """Stop animation of plot with objects using TimeSeries."""
        self.send({"msg_type": "stop_auto_play"})

    def close(self):
        """Remove plot from all its ipywidgets.Output()-s."""
        for output in self.outputs:
            output.clear_output()

        self.outputs = []

    def camera_reset(self, factor=1.5):
        """Trigger auto-adjustment of camera.

        Useful when self.camera_auto_fit == False."""
        self.send({"msg_type": "reset_camera", "factor": factor})

    def get_auto_grid(self):
        d = np.stack([o.get_bounding_box() for o in self.objects])

        return np.dstack(
            [np.min(d[:, 0::2], axis=0), np.max(d[:, 1::2], axis=0)]
        ).flatten()

    def get_auto_camera(self, factor=1.5, yaw=25, pitch=15, bounds=None):
        """ Compute the camera vector from the specified parameters. If `bounds`
        is not provided, then the algorithm will obtain it from the available
        meshes.
        """
        if bounds is None:
            bounds = self.get_auto_grid()
        center = (bounds[::2] + bounds[1::2]) / 2.0
        radius = 0.5 * np.sum(np.abs(bounds[::2] - bounds[1::2]) ** 2) ** 0.5
        cam_distance = radius * factor / np.sin(np.deg2rad(self.camera_fov / 2.0))

        x = np.sin(np.deg2rad(pitch)) * np.cos(np.deg2rad(yaw))
        y = np.sin(np.deg2rad(pitch)) * np.sin(np.deg2rad(yaw))
        z = np.cos(np.deg2rad(pitch))

        if pitch not in [0, 180]:
            up = [0, 0, 1]
        else:
            up = [0, 1, 1]

        return [
            center[0] + x * cam_distance,
            center[1] + y * cam_distance,
            center[2] + z * cam_distance,
            center[0],
            center[1],
            center[2],
            up[0],
            up[1],
            up[2],
        ]

    def fetch_screenshot(self, only_canvas=False):
        """Request creating a PNG screenshot on the JS side and saving it in self.screenshot

        The result is a string of a PNG file in base64 encoding.
        This function requires a round-trip of websocket messages. The result will
        be available after the current cell finishes execution."""
        self.send({"msg_type": "fetch_screenshot", "only_canvas": only_canvas})

    def yield_screenshots(self, generator_function):
        """Decorator for a generator function receiving screenshots via yield."""

        @wraps(generator_function)
        def inner():
            generator = generator_function()

            def send_new_value(change):
                try:
                    generator.send(base64.b64decode(change.new))
                except StopIteration:
                    self.unobserve(send_new_value, "screenshot")

            self.observe(send_new_value, "screenshot")
            # start the decorated generator
            generator.send(None)

        return inner

    def fetch_snapshot(self, compression_level=9):
        """Request creating a HTML snapshot on the JS side and saving it in self.snapshot

        The result is a string: a HTML document with this plot embedded.
        This function requires a round-trip of websocket messages. The result will
        be available after the current cell finishes execution."""
        self.send(
            {"msg_type": "fetch_snapshot", "compression_level": compression_level}
        )

    def yield_snapshots(self, generator_function):
        """Decorator for a generator function receiving snapshots via yield."""

        @wraps(generator_function)
        def inner():
            generator = generator_function()

            def send_new_value(change):
                try:
                    generator.send(base64.b64decode(change.new))
                except StopIteration:
                    self.unobserve(send_new_value, "snapshot")

            self.observe(send_new_value, "snapshot")
            # start the decorated generator
            generator.send(None)

        return inner

    def get_binary_snapshot(self, compression_level=9, voxel_chunks=[]):
        import zlib
        import msgpack

        snapshot = self.get_binary_snapshot_objects(voxel_chunks)
        snapshot['plot'] = self.get_plot_params()

        data = msgpack.packb(snapshot, use_bin_type=True)

        return zlib.compress(data, compression_level)

    def load_binary_snapshot(self, data):
        import zlib
        import msgpack
        from .helpers import from_json

        data = msgpack.unpackb(zlib.decompress(data))
        self.voxel_chunks = []

        for name in ['objects', 'chunkList']:
            if name in data.keys():
                for o in data[name]:
                    attributes = {
                        k: from_json(o[k]) for k in o.keys() if k != 'type'
                    }

                    if name == 'objects':
                        o = objects_map[o['type']](**attributes)
                        self += o
                    else:
                        self.voxel_chunks.append(VoxelChunk(**attributes))

        return self.voxel_chunks

    def get_binary_snapshot_objects(self, voxel_chunks=[]):
        from .helpers import to_json

        snapshot = {"objects": [], "chunkList": []}

        for name, l in [('objects', self.objects), ('chunkList', voxel_chunks)]:
            for o in l:
                obj = {}
                for k, v in o.traits().items():
                    if "sync" in v.metadata:
                        obj[k] = to_json(k, o[k], o, o["compression_level"])

                snapshot[name].append(obj)

        return snapshot

    def get_plot_params(self):
        return {
            "cameraAutoFit": self.camera_auto_fit,
            "viewMode": self.mode,
            "menuVisibility": self.menu_visibility,
            "gridAutoFit": self.grid_auto_fit,
            "gridVisible": self.grid_visible,
            "grid": self.grid,
            "gridColor": self.grid_color,
            "labelColor": self.label_color,
            "antialias": self.antialias,
            "screenshotScale": self.screenshot_scale,
            "clearColor": self.background_color,
            "clippingPlanes": self.clipping_planes,
            "lighting": self.lighting,
            "time": self.time,
            "fpsMeter": self.fps_meter,
            "cameraMode": self.camera_mode,
            "colorbarObjectId": self.colorbar_object_id,
            "axes": self.axes,
            "camera": self.camera,
            "cameraNoRotate": self.camera_no_rotate,
            "cameraNoZoom": self.camera_no_zoom,
            "cameraNoPan": self.camera_no_pan,
            "cameraRotateSpeed": self.camera_rotate_speed,
            "cameraZoomSpeed": self.camera_zoom_speed,
            "cameraPanSpeed": self.camera_pan_speed,
            "cameraDampingFactor": self.camera_damping_factor,
            "name": self.name,
            "height": self.height,
            "cameraFov": self.camera_fov,
            "axesHelper": self.axes_helper,
            "cameraAnimation": self.camera_animation,
            "fps": self.fps,
        }

    def get_snapshot(self, compression_level=9, voxel_chunks=[], additional_js_code=""):
        """Produce on the Python side a HTML document with the current plot embedded."""
        import os
        import io
        import zlib

        dir_path = os.path.dirname(os.path.realpath(__file__))

        data = self.get_binary_snapshot(compression_level, voxel_chunks)

        if self.snapshot_type == 'full':
            f = io.open(
                os.path.join(dir_path, "static", "snapshot_standalone.txt"),
                mode="r",
                encoding="utf-8",
            )
            template = f.read()
            f.close()

            f = io.open(
                os.path.join(dir_path, "static", "standalone.js"),
                mode="r",
                encoding="utf-8",
            )
            template = template.replace(
                "[K3D_SOURCE]",
                base64.b64encode(
                    zlib.compress(f.read().encode(), compression_level)
                ).decode("utf-8"),
            )
            f.close()

            f = io.open(
                os.path.join(dir_path, "static", "require.js"),
                mode="r",
                encoding="utf-8",
            )
            template = template.replace("[REQUIRE_JS]", f.read())
            f.close()

            f = io.open(
                os.path.join(dir_path, "static", "pako_inflate.min.js"),
                mode="r",
                encoding="utf-8",
            )
            template = template.replace("[PAKO_JS]", f.read())
            f.close()
        else:
            if self.snapshot_type == 'online':
                template_file = 'snapshot_online.txt'
            elif self.snapshot_type == 'inline':
                template_file = 'snapshot_inline.txt'
            else:
                raise Exception('Unknown snapshot_type')

            f = io.open(
                os.path.join(dir_path, "static", template_file),
                mode="r",
                encoding="utf-8",
            )
            template = f.read()
            f.close()

            template = template.replace("[VERSION]", self._view_module_version)
            template = template.replace("[HEIGHT]", str(self.height))
            template = template.replace("[ID]", str(id(self)))

        template = template.replace("[DATA]", base64.b64encode(data).decode("utf-8"))
        template = template.replace("[ADDITIONAL]", additional_js_code)

        return template

    def get_static_path(self):
        import os

        dir_path = os.path.dirname(os.path.realpath(__file__))

        return os.path.join(dir_path, "static")



import k3d
import numpy as np
import pytest
from .plot_compare import *

vertices = [
    -10.0, -15.0, 0.0,
    -10.0, 5.0, 0.0,
    10.0, 5.0, 0.0,
    -10.0, -15.0, 0.0,
    10.0, -15.0, 0.0,
    -10.0, 5.0, 0.0,
    0.0, 15.0, 0.0,
    10.0, 5.0, 0.0,
    10.0, -15.0, 0.0
]

colors = [
    0xff,
    0xffff,
    0xff00ff,
    0x00ffff,
    0xffff00,
    0xff00,
    0xff0000,
    0xff00ff,
    0xffffff
]


def test_line_simple_red():
    global vertices

    prepare()

    lines = k3d.line(vertices, color=0xff0000, shader="simple",
                     transform=k3d.transform(rotation=(np.pi / 2, 1, 0, 0)))
    pytest.plot += lines

    compare('line_simple_red')


def test_line_simple():
    global vertices, colors

    prepare()

    lines = k3d.line(vertices, colors=colors, shader="simple",
                     transform=k3d.transform(rotation=(np.pi / 2, 1, 0, 0)))
    pytest.plot += lines

    compare('line_simple')


def test_line_simple_clipping_plane():
    global vertices, colors

    prepare()
    pytest.plot.clipping_planes = [
        [1, 1, 0, 0]
    ]

    lines = k3d.line(vertices, colors=colors, shader="simple",
                     transform=k3d.transform(rotation=(np.pi / 2, 1, 0, 0)))
    pytest.plot += lines

    compare('line_simple_clipping_plane')


def test_line_thick():
    global vertices, colors

    prepare()

    lines = k3d.line(vertices, colors=colors, width=0.5, shader="thick",
                     transform=k3d.transform(rotation=(np.pi / 2, 1, 0, 0)))
    pytest.plot += lines

    compare('line_thick')


def test_line_thick_clipping_plane():
    global vertices, colors

    prepare()
    pytest.plot.clipping_planes = [
        [1, 1, 0, 0]
    ]
    lines = k3d.line(vertices, colors=colors, width=0.5, shader="thick",
                     transform=k3d.transform(rotation=(np.pi / 2, 1, 0, 0)))
    pytest.plot += lines

    compare('line_thick_clipping_plane')


def test_line_mesh():
    global vertices, colors

    prepare()

    lines = k3d.line(vertices, colors=colors, width=0.5, shader="mesh", radial_segments=8,
                     transform=k3d.transform(rotation=(np.pi / 2, 1, 0, 0)))
    pytest.plot += lines

    compare('line_mesh')


def test_line_mesh_clipping_plane():
    global vertices, colors

    prepare()
    pytest.plot.clipping_planes = [
        [1, 1, 0, 0]
    ]
    lines = k3d.line(vertices, colors=colors, width=0.5, shader="mesh", radial_segments=8,
                     transform=k3d.transform(rotation=(np.pi / 2, 1, 0, 0)))
    pytest.plot += lines

    compare('line_mesh_clipping_plane')


def test_line_simplified_mesh():
    global vertices, colors

    prepare()

    lines = k3d.line(vertices, colors=colors, width=0.5, shader="mesh", radial_segments=3,
                     transform=k3d.transform(rotation=(np.pi / 2, 1, 0, 0)))
    pytest.plot += lines

    compare('line_simplified_mesh')


def test_line_advanced():
    import math

    prepare()

    for i in range(10):
        v = []
        s = []

        for fi in np.arange(0, 2 * math.pi, 0.01):
            v.append([
                math.sin(fi * 4 + i),
                math.cos(fi * 7 + i),
                math.cos(fi * 3) + fi * 1.5
            ])
            s.append(math.sin(fi * i))

        pytest.plot += k3d.line(v, attribute=s, color_range=[-1, 1],
                                color_map=k3d.basic_color_maps.Jet, width=0.03)

    compare('line_advanced')
